import logging
import os
from argparse import Action, ArgumentParser, _AppendAction
from typing import TYPE_CHECKING, Any, Callable, Optional, Union

from configargparse import SUPPRESS
from ruamel.yaml import YAML

from toil.batchSystems.options import add_all_batchsystem_options
from toil.lib.conversions import bytes2human, human2bytes, opt_strtobool, strtobool
from toil.provisioners import parse_node_types
from toil.statsAndLogging import add_logging_options

if TYPE_CHECKING:
    from toil.job import AcceleratorRequirement

logger = logging.getLogger(__name__)

# aim to pack autoscaling jobs within a 30 minute block before provisioning a new node
defaultTargetTime = 1800
SYS_MAX_SIZE = 9223372036854775807
# sys.max_size on 64 bit systems is 9223372036854775807, so that 32-bit systems
# use the same number


def parse_set_env(l: list[str]) -> dict[str, Optional[str]]:
    """
    Parse a list of strings of the form "NAME=VALUE" or just "NAME" into a dictionary.

    Strings of the latter from will result in dictionary entries whose value is None.

    >>> parse_set_env([])
    {}
    >>> parse_set_env(['a'])
    {'a': None}
    >>> parse_set_env(['a='])
    {'a': ''}
    >>> parse_set_env(['a=b'])
    {'a': 'b'}
    >>> parse_set_env(['a=a', 'a=b'])
    {'a': 'b'}
    >>> parse_set_env(['a=b', 'c=d'])
    {'a': 'b', 'c': 'd'}
    >>> parse_set_env(['a=b=c'])
    {'a': 'b=c'}
    >>> parse_set_env([''])
    Traceback (most recent call last):
    ...
    ValueError: Empty name
    >>> parse_set_env(['=1'])
    Traceback (most recent call last):
    ...
    ValueError: Empty name
    """
    d = {}
    v: Optional[str] = None
    for i in l:
        try:
            k, v = i.split("=", 1)
        except ValueError:
            k, v = i, None
        if not k:
            raise ValueError("Empty name")
        d[k] = v
    return d


def parse_str_list(s: str) -> list[str]:
    return [str(x) for x in s.split(",")]


def parse_int_list(s: str) -> list[int]:
    return [int(x) for x in s.split(",")]


def iC(min_value: int, max_value: Optional[int] = None) -> Callable[[int], bool]:
    """Returns a function that checks if a given int is in the given half-open interval."""
    assert isinstance(min_value, int)
    if max_value is None:
        return lambda x: min_value <= x
    assert isinstance(max_value, int)
    return lambda x: min_value <= x < max_value


def fC(minValue: float, maxValue: Optional[float] = None) -> Callable[[float], bool]:
    """Returns a function that checks if a given float is in the given half-open interval."""
    assert isinstance(minValue, float)
    if maxValue is None:
        return lambda x: minValue <= x
    assert isinstance(maxValue, float)
    return lambda x: minValue <= x < maxValue


def parse_accelerator_list(specs: Optional[str]) -> list["AcceleratorRequirement"]:
    """
    Parse a string description of one or more accelerator requirements.
    """

    if specs is None or len(specs) == 0:
        # Not specified, so the default default is to not need any.
        return []
    # Otherwise parse each requirement.
    from toil.job import parse_accelerator

    return [parse_accelerator(r) for r in specs.split(",")]


def parseBool(val: str) -> bool:
    if val.lower() in ["true", "t", "yes", "y", "on", "1"]:
        return True
    elif val.lower() in ["false", "f", "no", "n", "off", "0"]:
        return False
    else:
        raise RuntimeError('Could not interpret "%s" as a boolean value' % val)


# This is kept in the outer scope as multiple batchsystem files use this
def make_open_interval_action(
    min: Union[int, float], max: Optional[Union[int, float]] = None
) -> type[Action]:
    """
    Returns an argparse action class to check if the input is within the given half-open interval.
    ex:
    Provided value to argparse must be within the interval [min, max)
    Types of min and max must be the same (max may be None)

    :param min: float/int
    :param max: optional float/int
    :return: argparse action class
    """

    class IntOrFloatOpenAction(Action):
        def __call__(
            self, parser: Any, namespace: Any, values: Any, option_string: Any = None
        ) -> None:
            if isinstance(min, int):
                if max is not None:  # for mypy
                    assert isinstance(max, int)
                func = iC(min, max)
            else:
                func = fC(min, max)
            try:
                if not func(values):
                    if max is None:
                        raise parser.error(
                            f"{option_string} ({values}) must be at least {min}"
                        )
                    else:
                        raise parser.error(
                            f"{option_string} ({values}) must be at least {min} and strictly less than {max})"
                        )
            except AssertionError:
                raise RuntimeError(
                    f"The {option_string} option has an invalid value: {values}"
                )
            setattr(namespace, self.dest, values)

    return IntOrFloatOpenAction


def parse_jobstore(jobstore_uri: str) -> str:
    """
    Turn the jobstore string into it's corresponding URI
    ex:
    /path/to/jobstore -> file:/path/to/jobstore

    If the jobstore string already is a URI, return the jobstore:
    aws:/path/to/jobstore -> aws:/path/to/jobstore
    :param jobstore_uri: string of the jobstore
    :return: URI of the jobstore
    """
    from toil.common import Toil

    name, rest = Toil.parseLocator(jobstore_uri)
    if name == "file":
        # We need to resolve relative paths early, on the leader, because the worker process
        # may have a different working directory than the leader, e.g. under Mesos.
        return Toil.buildLocator(name, os.path.abspath(rest))
    else:
        return jobstore_uri


JOBSTORE_HELP = (
    "The location of the job store for the workflow.  "
    "A job store holds persistent information about the jobs, stats, and files in a "
    "workflow. If the workflow is run with a distributed batch system, the job "
    "store must be accessible by all worker nodes. Depending on the desired "
    "job store implementation, the location should be formatted according to "
    "one of the following schemes:\n\n"
    "file:<path> where <path> points to a directory on the file system\n\n"
    "aws:<region>:<prefix> where <region> is the name of an AWS region like "
    "us-west-2 and <prefix> will be prepended to the names of any top-level "
    "AWS resources in use by job store, e.g. S3 buckets.\n\n "
    "google:<project_id>:<prefix> TODO: explain\n\n"
    "For backwards compatibility, you may also specify ./foo (equivalent to "
    "file:./foo or just file:foo) or /bar (equivalent to file:/bar)."
)


def add_base_toil_options(
    parser: ArgumentParser, jobstore_as_flag: bool = False, cwl: bool = False
) -> None:
    """
    Add base Toil command line options to the parser.
    :param parser: Argument parser to add options to
    :param jobstore_as_flag: make the job store option a --jobStore flag instead of a required jobStore positional argument.
    :param cwl: whether CWL should be included or not
    """

    # This is necessary as the config file must have at least one valid key to parse properly and we want to use a
    # dummy key
    config = parser.add_argument_group()
    config.add_argument("--config_version", default=None, help=SUPPRESS)

    # If using argparse instead of configargparse, this should just not parse when calling parse_args()
    # default config value is set to none as defaults should already be populated at config init
    config.add_argument(
        "--config",
        dest="config",
        is_config_file_arg=True,
        default=None,
        metavar="PATH",
        help="Get options from a config file.",
    )

    add_logging_options(parser)
    parser.register("type", "bool", parseBool)  # Custom type for arg=True/False.

    # Core options
    core_options = parser.add_argument_group(
        title="Toil core options",
        description="Options to specify the location of the Toil workflow and "
        "turn on stats collation about the performance of jobs.",
    )
    if jobstore_as_flag:
        core_options.add_argument(
            "--jobstore",
            "--jobStore",
            dest="jobStore",
            type=parse_jobstore,
            default=None,
            help=JOBSTORE_HELP,
        )
    else:
        core_options.add_argument("jobStore", type=parse_jobstore, help=JOBSTORE_HELP)

    class WorkDirAction(Action):
        """
        Argparse action class to check that the provided --workDir exists
        """

        def __call__(
            self, parser: Any, namespace: Any, values: Any, option_string: Any = None
        ) -> None:
            workDir = values
            if workDir is not None:
                workDir = os.path.abspath(workDir)
                if not os.path.exists(workDir):
                    raise RuntimeError(
                        f"The path provided to --workDir ({workDir}) does not exist."
                    )

                if len(workDir) > 80:
                    logger.warning(
                        f'Length of workDir path "{workDir}" is {len(workDir)} characters.  '
                        f"Consider setting a shorter path with --workPath or setting TMPDIR to something "
                        f'like "/tmp" to avoid overly long paths.'
                    )
            setattr(namespace, self.dest, workDir)

    class CoordinationDirAction(Action):
        """
        Argparse action class to check that the provided --coordinationDir exists
        """

        def __call__(
            self, parser: Any, namespace: Any, values: Any, option_string: Any = None
        ) -> None:
            coordination_dir = values
            if coordination_dir is not None:
                coordination_dir = os.path.abspath(coordination_dir)
                if not os.path.exists(coordination_dir):
                    raise RuntimeError(
                        f"The path provided to --coordinationDir ({coordination_dir}) does not exist."
                    )
            setattr(namespace, self.dest, coordination_dir)

    def make_closed_interval_action(
        min: Union[int, float], max: Optional[Union[int, float]] = None
    ) -> type[Action]:
        """
        Returns an argparse action class to check if the input is within the given half-open interval.
        ex:
        Provided value to argparse must be within the interval [min, max]

        :param min: int/float
        :param max: optional int/float
        :return: argparse action
        """

        class ClosedIntOrFloatAction(Action):
            def __call__(
                self,
                parser: Any,
                namespace: Any,
                values: Any,
                option_string: Any = None,
            ) -> None:
                def is_within(x: Union[int, float]) -> bool:
                    if max is None:
                        return min <= x
                    else:
                        return min <= x <= max

                try:
                    if not is_within(values):
                        raise parser.error(
                            f"{option_string} ({values}) must be within the range: [{min}, {'infinity' if max is None else max}]"
                        )
                except AssertionError:
                    raise RuntimeError(
                        f"The {option_string} option has an invalid value: {values}"
                    )
                setattr(namespace, self.dest, values)

        return ClosedIntOrFloatAction

    core_options.add_argument(
        "--workDir",
        dest="workDir",
        default=None,
        env_var="TOIL_WORKDIR",
        action=WorkDirAction,
        metavar="PATH",
        help="Absolute path to directory where temporary files generated during the Toil "
        "run should be placed. Standard output and error from batch system jobs "
        "(unless --noStdOutErr is set) will be placed in this directory. A cache directory "
        "may be placed in this directory. Temp files and folders will be placed in a "
        "directory toil-<workflowID> within workDir. The workflowID is generated by "
        "Toil and will be reported in the workflow logs. Default is determined by the "
        "variables (TMPDIR, TEMP, TMP) via mkdtemp. This directory needs to exist on "
        "all machines running jobs; if capturing standard output and error from batch "
        "system jobs is desired, it will generally need to be on a shared file system. "
        "When sharing a cache between containers on a host, this directory must be "
        "shared between the containers.",
    )
    core_options.add_argument(
        "--coordinationDir",
        dest="coordination_dir",
        default=None,
        env_var="TOIL_COORDINATION_DIR",
        action=CoordinationDirAction,
        metavar="PATH",
        help="Absolute path to directory where Toil will keep state and lock files. "
        "When sharing a cache between containers on a host, this directory must be "
        "shared between the containers.",
    )
    core_options.add_argument(
        "--noStdOutErr",
        dest="noStdOutErr",
        default=False,
        action="store_true",
        help="Do not capture standard output and error from batch system jobs.",
    )
    # TODO: Should this be deprecated since we always save stats now for history tracking?
    core_options.add_argument(
        "--stats",
        dest="stats",
        default=False,
        action="store_true",
        help="Keep statistics about the toil workflow to be used by 'toil stats'.",
    )
    clean_choices = ["always", "onError", "never", "onSuccess"]
    core_options.add_argument(
        "--clean",
        dest="clean",
        choices=clean_choices,
        default="onSuccess",
        help=f"Determines the deletion of the jobStore upon completion of the program.  "
        f"Choices: {clean_choices}.  The --stats option requires information from the "
        f"jobStore upon completion so the jobStore will never be deleted with that flag.  "
        f"If you wish to be able to restart the run, choose 'never' or 'onSuccess'.  "
        f"Default is 'never' if stats is enabled, and 'onSuccess' otherwise.",
    )
    core_options.add_argument(
        "--cleanWorkDir",
        dest="cleanWorkDir",
        choices=clean_choices,
        default="always",
        help=f"Determines deletion of temporary worker directory upon completion of a job.  "
        f"Choices: {clean_choices}.  Default = always.  WARNING: This option should be "
        f"changed for debugging only.  Running a full pipeline with this option could "
        f"fill your disk with excessive intermediate data.",
    )
    core_options.add_argument(
        "--clusterStats",
        dest="clusterStats",
        nargs="?",
        action="store",
        default=None,
        metavar="OPT_PATH",
        const=os.getcwd(),
        help="If enabled, writes out JSON resource usage statistics to a file.  "
        "The default location for this file is the current working directory, but an "
        "absolute path can also be passed to specify where this file should be written. "
        "This options only applies when using scalable batch systems.",
    )

    # Restarting the workflow options
    restart_options = parser.add_argument_group(
        title="Toil options for restarting an existing workflow",
        description="Allows the restart of an existing workflow",
    )
    restart_options.add_argument(
        "--restart",
        dest="restart",
        default=False,
        action="store_true",
        help="If --restart is specified then will attempt to restart existing workflow "
        "at the location pointed to by the --jobStore option. Will raise an exception "
        "if the workflow does not exist",
    )

    # Batch system options
    batchsystem_options = parser.add_argument_group(
        title="Toil options for specifying the batch system",
        description="Allows the specification of the batch system.",
    )
    add_all_batchsystem_options(batchsystem_options)

    # File store options
    file_store_options = parser.add_argument_group(
        title="Toil options for configuring storage",
        description="Allows configuring Toil's data storage.",
    )
    link_imports = file_store_options.add_mutually_exclusive_group()
    link_imports_help = (
        "When using a filesystem based job store, CWL input files are by default symlinked in.  "
        "Setting this option to True instead copies the files into the job store, which may protect "
        "them from being modified externally.  When set to False, as long as caching is enabled, "
        "Toil will protect the file automatically by changing the permissions to read-only. "
        "default=%(default)s"
    )
    link_imports.add_argument(
        "--symlinkImports",
        dest="symlinkImports",
        type=strtobool,
        default=True,
        metavar="BOOL",
        help=link_imports_help,
    )
    move_exports = file_store_options.add_mutually_exclusive_group()
    move_exports_help = (
        "When using a filesystem based job store, output files are by default moved to the "
        "output directory, and a symlink to the moved exported file is created at the initial "
        "location.  Setting this option to True instead copies the files into the output directory.  "
        "Applies to filesystem-based job stores only. "
        "default=%(default)s"
    )
    move_exports.add_argument(
        "--moveOutputs",
        dest="moveOutputs",
        type=strtobool,
        default=False,
        metavar="BOOL",
        help=move_exports_help,
    )

    caching = file_store_options.add_mutually_exclusive_group()
    caching_help = ("Enable or disable worker level file caching for your workflow, specifying this overrides default from batch system. "
                    "Does not affect CWL or WDL task caching.")
    caching.add_argument(
        "--caching",
        dest="caching",
        type=opt_strtobool,
        default=None,
        metavar="BOOL",
        help=caching_help,
    )
    # default is None according to PR 4299, seems to be generated at runtime

    file_store_options.add_argument(
        "--symlinkJobStoreReads",
        dest="symlink_job_store_reads",
        type=strtobool,
        default=True,
        metavar="BOOL",
        help="Allow reads and container mounts from a JobStore's shared filesystem directly "
        "via symlink. default=%(default)s",
    )

    # Auto scaling options
    autoscaling_options = parser.add_argument_group(
        title="Toil options for autoscaling the cluster of worker nodes",
        description="Allows the specification of the minimum and maximum number of nodes in an autoscaled cluster, "
        "as well as parameters to control the level of provisioning.",
    )
    provisioner_choices = ["aws", "gce", None]

    # TODO: Better consolidate this provisioner arg and the one in provisioners/__init__.py?
    autoscaling_options.add_argument(
        "--provisioner",
        "-p",
        dest="provisioner",
        choices=provisioner_choices,
        default=None,
        help=f"The provisioner for cluster auto-scaling.  This is the main Toil "
        f"'--provisioner' option, and defaults to None for running on single "
        f"machine and non-auto-scaling batch systems.  The currently supported "
        f"choices are {provisioner_choices}.  The default is %(default)s.",
    )
    autoscaling_options.add_argument(
        "--nodeTypes",
        default=[],
        dest="nodeTypes",
        type=parse_node_types,
        action="extend",
        help="Specifies a list of comma-separated node types, each of which is "
        "composed of slash-separated instance types, and an optional spot "
        "bid set off by a colon, making the node type preemptible. Instance "
        "types may appear in multiple node types, and the same node type "
        "may appear as both preemptible and non-preemptible.\n"
        "Valid argument specifying two node types:\n"
        "\tc5.4xlarge/c5a.4xlarge:0.42,t2.large\n"
        "Node types:\n"
        "\tc5.4xlarge/c5a.4xlarge:0.42 and t2.large\n"
        "Instance types:\n"
        "\tc5.4xlarge, c5a.4xlarge, and t2.large\n"
        "Semantics:\n"
        "\tBid $0.42/hour for either c5.4xlarge or c5a.4xlarge instances,\n"
        "\ttreated interchangeably, while they are available at that price,\n"
        "\tand buy t2.large instances at full price.\n"
        "default=%(default)s",
    )

    class NodeExtendAction(_AppendAction):
        """
        argparse Action class to remove the default value on first call, and act as an extend action after
        """

        # with action=append/extend, the argparse default is always prepended to the option
        # so make the CLI have priority by rewriting the option on the first run
        def __init__(self, option_strings: Any, dest: Any, **kwargs: Any):
            super().__init__(option_strings, dest, **kwargs)
            self.is_default = True

        def __call__(
            self, parser: Any, namespace: Any, values: Any, option_string: Any = None
        ) -> None:
            if self.is_default:
                setattr(namespace, self.dest, values)
                self.is_default = False
            else:
                super().__call__(parser, namespace, values, option_string)

    autoscaling_options.add_argument(
        "--maxNodes",
        default=[10],
        dest="maxNodes",
        type=parse_int_list,
        action=NodeExtendAction,
        metavar="INT[,INT...]",
        help=f"Maximum number of nodes of each type in the cluster, if using autoscaling, "
        f"provided as a comma-separated list.  The first value is used as a default "
        f"if the list length is less than the number of nodeTypes.  "
        f"default=%(default)s",
    )
    autoscaling_options.add_argument(
        "--minNodes",
        default=[0],
        dest="minNodes",
        type=parse_int_list,
        action=NodeExtendAction,
        metavar="INT[,INT...]",
        help="Mininum number of nodes of each type in the cluster, if using "
        "auto-scaling.  This should be provided as a comma-separated list of the "
        "same length as the list of node types. default=%(default)s",
    )
    autoscaling_options.add_argument(
        "--targetTime",
        dest="targetTime",
        default=defaultTargetTime,
        type=int,
        action=make_closed_interval_action(0),
        metavar="INT",
        help=f"Sets how rapidly you aim to complete jobs in seconds. Shorter times mean "
        f"more aggressive parallelization. The autoscaler attempts to scale up/down "
        f"so that it expects all queued jobs will complete within targetTime "
        f"seconds.  default=%(default)s",
    )
    autoscaling_options.add_argument(
        "--betaInertia",
        dest="betaInertia",
        default=0.1,
        type=float,
        action=make_closed_interval_action(0.0, 0.9),
        metavar="FLOAT",
        help=f"A smoothing parameter to prevent unnecessary oscillations in the number "
        f"of provisioned nodes. This controls an exponentially weighted moving "
        f"average of the estimated number of nodes. A value of 0.0 disables any "
        f"smoothing, and a value of 0.9 will smooth so much that few changes will "
        f"ever be made.  Must be between 0.0 and 0.9. default=%(default)s",
    )
    autoscaling_options.add_argument(
        "--scaleInterval",
        dest="scaleInterval",
        default=60,
        type=int,
        metavar="INT",
        help=f"The interval (seconds) between assessing if the scale of "
        f"the cluster needs to change. default=%(default)s",
    )
    autoscaling_options.add_argument(
        "--preemptibleCompensation",
        "--preemptableCompensation",
        dest="preemptibleCompensation",
        default=0.0,
        type=float,
        action=make_closed_interval_action(0.0, 1.0),
        metavar="FLOAT",
        help=f"The preference of the autoscaler to replace preemptible nodes with "
        f"non-preemptible nodes, when preemptible nodes cannot be started for some "
        f"reason. This value must be between 0.0 and 1.0, inclusive.  "
        f"A value of 0.0 disables such "
        f"compensation, a value of 0.5 compensates two missing preemptible nodes "
        f"with a non-preemptible one. A value of 1.0 replaces every missing "
        f"pre-emptable node with a non-preemptible one. default=%(default)s",
    )
    autoscaling_options.add_argument(
        "--nodeStorage",
        dest="nodeStorage",
        default=50,
        type=int,
        metavar="INT",
        help="Specify the size of the root volume of worker nodes when they are launched "
        "in gigabytes. You may want to set this if your jobs require a lot of disk "
        f"space.  (default=%(default)s).",
    )
    autoscaling_options.add_argument(
        "--nodeStorageOverrides",
        dest="nodeStorageOverrides",
        default=[],
        type=parse_str_list,
        action="extend",
        metavar="NODETYPE:NODESTORAGE[,NODETYPE:NODESTORAGE...]",
        help="Comma-separated list of nodeType:nodeStorage that are used to override "
        "the default value from --nodeStorage for the specified nodeType(s).  "
        "This is useful for heterogeneous jobs where some tasks require much more "
        "disk than others.",
    )

    autoscaling_options.add_argument(
        "--metrics",
        dest="metrics",
        default=False,
        type=strtobool,
        metavar="BOOL",
        help="Enable the prometheus/grafana dashboard for monitoring CPU/RAM usage, "
        "queue size, and issued jobs.",
    )
    autoscaling_options.add_argument(
        "--assumeZeroOverhead",
        dest="assume_zero_overhead",
        default=False,
        type=strtobool,
        metavar="BOOL",
        help="Ignore scheduler and OS overhead and assume jobs can use every last byte "
        "of memory and disk on a node when autoscaling.",
    )

    # Parameters to limit service jobs / detect service deadlocks
    service_options = parser.add_argument_group(
        title="Toil options for limiting the number of service jobs and detecting service deadlocks",
        description=(
            SUPPRESS
            if cwl
            else "Allows the specification of the maximum number of service jobs in a cluster. "
            "By keeping this limited we can avoid nodes occupied with services causing "
            "deadlocks."
        ),
    )
    service_options.add_argument(
        "--maxServiceJobs",
        dest="maxServiceJobs",
        default=SYS_MAX_SIZE,
        type=int,
        metavar="INT",
        help=(
            SUPPRESS
            if cwl
            else f"The maximum number of service jobs that can be run "
            f"concurrently, excluding service jobs running on "
            f"preemptible nodes.  default=%(default)s"
        ),
    )
    service_options.add_argument(
        "--maxPreemptibleServiceJobs",
        dest="maxPreemptibleServiceJobs",
        default=SYS_MAX_SIZE,
        type=int,
        metavar="INT",
        help=(
            SUPPRESS
            if cwl
            else "The maximum number of service jobs that can run "
            "concurrently on preemptible nodes.  default=%(default)s"
        ),
    )
    service_options.add_argument(
        "--deadlockWait",
        dest="deadlockWait",
        default=60,
        type=int,
        metavar="INT",
        help=(
            SUPPRESS
            if cwl
            else f"Time, in seconds, to tolerate the workflow running only "
            f"the same service jobs, with no jobs to use them, "
            f"before declaring the workflow to be deadlocked and "
            f"stopping.  default=%(default)s"
        ),
    )
    service_options.add_argument(
        "--deadlockCheckInterval",
        dest="deadlockCheckInterval",
        default=30,
        type=int,
        metavar="INT",
        help=(
            SUPPRESS
            if cwl
            else "Time, in seconds, to wait between checks to see if the "
            "workflow is stuck running only service jobs, with no jobs "
            "to use them. Should be shorter than --deadlockWait. May "
            "need to be increased if the batch system cannot enumerate "
            "running jobs quickly enough, or if polling for running "
            "jobs is placing an unacceptable load on a shared cluster."
            f"default=%(default)s"
        ),
    )

    # Resource requirements
    resource_options = parser.add_argument_group(
        title="Toil options for cores/memory requirements",
        description="The options to specify default cores/memory requirements (if not specified by the jobs "
        "themselves), and to limit the total amount of memory/cores requested from the batch system.",
    )
    resource_help_msg = (
        "The {} amount of {} to request for a job.  "
        "Only applicable to jobs that do not specify an explicit value for this requirement.  "
        "{}.  "
        "Default is {}."
    )
    cpu_note = "Fractions of a core (for example 0.1) are supported on some batch systems [mesos, single_machine]"
    disk_mem_note = "Standard suffixes like K, Ki, M, Mi, G or Gi are supported"
    accelerators_note = (
        "Each accelerator specification can have a type (gpu [default], nvidia, amd, cuda, rocm, opencl, "
        "or a specific model like nvidia-tesla-k80), and a count [default: 1]. If both a type and a count "
        "are used, they must be separated by a colon. If multiple types of accelerators are "
        "used, the specifications are separated by commas"
    )

    h2b = lambda x: human2bytes(str(x))

    resource_options.add_argument(
        "--defaultMemory",
        dest="defaultMemory",
        default="2.0 Gi",
        type=h2b,
        action=make_open_interval_action(1),
        help=resource_help_msg.format(
            "default", "memory", disk_mem_note, bytes2human(2147483648)
        ),
    )
    resource_options.add_argument(
        "--defaultCores",
        dest="defaultCores",
        default=1,
        metavar="FLOAT",
        type=float,
        action=make_open_interval_action(1.0),
        help=resource_help_msg.format("default", "cpu", cpu_note, str(1)),
    )
    resource_options.add_argument(
        "--defaultDisk",
        dest="defaultDisk",
        default="2.0 Gi",
        metavar="INT",
        type=h2b,
        action=make_open_interval_action(1),
        help=resource_help_msg.format(
            "default", "disk", disk_mem_note, bytes2human(2147483648)
        ),
    )
    resource_options.add_argument(
        "--defaultAccelerators",
        dest="defaultAccelerators",
        default=[],
        metavar="ACCELERATOR[,ACCELERATOR...]",
        type=parse_accelerator_list,
        action="extend",
        help=resource_help_msg.format("default", "accelerators", accelerators_note, []),
    )
    resource_options.add_argument(
        "--defaultPreemptible",
        "--defaultPreemptable",
        dest="defaultPreemptible",
        metavar="BOOL",
        type=strtobool,
        nargs="?",
        const=True,
        default=False,
        help="Make all jobs able to run on preemptible (spot) nodes by default.",
    )
    resource_options.add_argument(
        "--maxCores",
        dest="maxCores",
        default=SYS_MAX_SIZE,
        metavar="INT",
        type=int,
        action=make_open_interval_action(1),
        help=resource_help_msg.format("max", "cpu", cpu_note, str(SYS_MAX_SIZE)),
    )
    resource_options.add_argument(
        "--maxMemory",
        dest="maxMemory",
        default=SYS_MAX_SIZE,
        metavar="INT",
        type=h2b,
        action=make_open_interval_action(1),
        help=resource_help_msg.format(
            "max", "memory", disk_mem_note, bytes2human(SYS_MAX_SIZE)
        ),
    )
    resource_options.add_argument(
        "--maxDisk",
        dest="maxDisk",
        default=SYS_MAX_SIZE,
        metavar="INT",
        type=h2b,
        action=make_open_interval_action(1),
        help=resource_help_msg.format(
            "max", "disk", disk_mem_note, bytes2human(SYS_MAX_SIZE)
        ),
    )

    # Retrying/rescuing jobs
    job_options = parser.add_argument_group(
        title="Toil options for rescuing/killing/restarting jobs",
        description="The options for jobs that either run too long/fail or get lost (some batch systems have issues!).",
    )
    job_options.add_argument(
        "--retryCount",
        dest="retryCount",
        default=1,
        type=int,
        action=make_open_interval_action(0),
        metavar="INT",
        help=f"Number of times to retry a failing job before giving up and "
        f"labeling job failed. default={1}",
    )
    job_options.add_argument(
        "--stopOnFirstFailure",
        dest="stop_on_first_failure",
        type=strtobool,
        default=False,
        metavar="BOOL",
        help="Stop the workflow at the first complete job failure.",
    )
    job_options.add_argument(
        "--enableUnlimitedPreemptibleRetries",
        "--enableUnlimitedPreemptableRetries",
        dest="enableUnlimitedPreemptibleRetries",
        type=strtobool,
        default=False,
        metavar="BOOL",
        help="If set, preemptible failures (or any failure due to an instance getting "
        "unexpectedly terminated) will not count towards job failures and --retryCount.",
    )
    job_options.add_argument(
        "--doubleMem",
        dest="doubleMem",
        type=strtobool,
        default=False,
        metavar="BOOL",
        help="If set, batch jobs which die to reaching memory limit on batch schedulers "
        "will have their memory doubled and they will be retried. The remaining "
        "retry count will be reduced by 1. Currently supported by LSF.",
    )
    job_options.add_argument(
        "--maxJobDuration",
        dest="maxJobDuration",
        default=SYS_MAX_SIZE,
        type=int,
        action=make_open_interval_action(1),
        metavar="INT",
        help=f"Maximum runtime of a job (in seconds) before we kill it (this is a lower bound, "
        f"and the actual time before killing the job may be longer).  "
        f"default=%(default)s",
    )
    job_options.add_argument(
        "--rescueJobsFrequency",
        dest="rescueJobsFrequency",
        default=60,
        type=int,
        action=make_open_interval_action(1),
        metavar="INT",
        help=f"Period of time to wait (in seconds) between checking for missing/overlong jobs, "
        f"that is jobs which get lost by the batch system. Expert parameter.  "
        f"default=%(default)s",
    )
    job_options.add_argument(
        "--jobStoreTimeout",
        dest="job_store_timeout",
        default=30,
        type=float,
        action=make_open_interval_action(0),
        metavar="FLOAT",
        help=f"Maximum time (in seconds) to wait for a job's update to the job store "
        f"before declaring it failed. default=%(default)s",
    )

    # Log management options
    log_options = parser.add_argument_group(
        title="Toil log management options",
        description="Options for how Toil should manage its logs.",
    )
    log_options.add_argument(
        "--maxLogFileSize",
        dest="maxLogFileSize",
        default=100 * 1024 * 1024,
        type=h2b,
        action=make_open_interval_action(1),
        help=f"The maximum size of a job log file to keep (in bytes), log files larger than "
        f"this will be truncated to the last X bytes. Setting this option to zero will "
        f"prevent any truncation. Setting this option to a negative value will truncate "
        f"from the beginning.  Default={bytes2human(100 * 1024 * 1024)}",
    )
    log_options.add_argument(
        "--writeLogs",
        dest="writeLogs",
        nargs="?",
        action="store",
        default=None,
        const=os.getcwd(),
        metavar="OPT_PATH",
        help="Write worker logs received by the leader into their own files at the specified "
        "path. Any non-empty standard output and error from failed batch system jobs will "
        "also be written into files at this path.  The current working directory will be "
        "used if a path is not specified explicitly. Note: By default only the logs of "
        "failed jobs are returned to leader. Set log level to 'debug' or enable "
        "'--writeLogsFromAllJobs' to get logs back from successful jobs, and adjust "
        "'maxLogFileSize' to control the truncation limit for worker logs.",
    )
    log_options.add_argument(
        "--writeLogsGzip",
        dest="writeLogsGzip",
        nargs="?",
        action="store",
        default=None,
        const=os.getcwd(),
        metavar="OPT_PATH",
        help="Identical to --writeLogs except the logs files are gzipped on the leader.",
    )
    log_options.add_argument(
        "--writeLogsFromAllJobs",
        dest="writeLogsFromAllJobs",
        type=strtobool,
        default=False,
        metavar="BOOL",
        help="Whether to write logs from all jobs (including the successful ones) without "
        "necessarily setting the log level to 'debug'. Ensure that either --writeLogs "
        "or --writeLogsGzip is set if enabling this option.",
    )
    log_options.add_argument(
        "--writeMessages",
        dest="write_messages",
        default=None,
        type=lambda x: None if x is None else os.path.abspath(x),
        metavar="PATH",
        help="File to send messages from the leader's message bus to.",
    )
    log_options.add_argument(
        "--realTimeLogging",
        dest="realTimeLogging",
        type=strtobool,
        default=False,
        metavar="BOOL",
        help="Enable real-time logging from workers to leader",
    )

    # Misc options
    misc_options = parser.add_argument_group(
        title="Toil miscellaneous options", description="Everything else."
    )
    misc_options.add_argument(
        "--disableChaining",
        dest="disableChaining",
        type=strtobool,
        default=False,
        metavar="BOOL",
        help="Disables chaining of jobs (chaining uses one job's resource allocation "
        "for its successor job if possible).",
    )
    misc_options.add_argument(
        "--disableJobStoreChecksumVerification",
        dest="disableJobStoreChecksumVerification",
        default=False,
        type=strtobool,
        metavar="BOOL",
        help="Disables checksum verification for files transferred to/from the job store.  "
        "Checksum verification is a safety check to ensure the data is not corrupted "
        "during transfer. Currently only supported for non-streaming AWS files.",
    )

    class SSEKeyAction(Action):
        def __call__(
            self, parser: Any, namespace: Any, values: Any, option_string: Any = None
        ) -> None:
            if values is not None:
                sse_key = values
                if sse_key is None:
                    return
                with open(sse_key) as f:
                    assert (
                        len(f.readline().rstrip()) == 32
                    ), "SSE key appears to be invalid."
            setattr(namespace, self.dest, values)

    misc_options.add_argument(
        "--sseKey",
        dest="sseKey",
        default=None,
        action=SSEKeyAction,
        metavar="PATH",
        help="Path to file containing 32 character key to be used for server-side encryption on "
        "awsJobStore or googleJobStore. SSE will not be used if this flag is not passed.",
    )

    # yaml.safe_load is being deprecated, this is the suggested workaround
    def yaml_safe_load(stream: Any) -> Any:
        yaml = YAML(typ="safe", pure=True)
        d = yaml.load(stream)
        if isinstance(d, dict):
            # this means the argument was a dictionary and is valid yaml (for configargparse)
            return d
        else:
            # this means the argument is likely in it's string format (for CLI)
            return parse_set_env(parse_str_list(stream))

    class ExtendActionDict(Action):
        """
        Argparse action class to implement the action="extend" functionality on dictionaries
        """

        def __call__(
            self, parser: Any, namespace: Any, values: Any, option_string: Any = None
        ) -> None:
            items = getattr(namespace, self.dest, None)
            assert (
                items is not None
            )  # for mypy. This should never be None, esp. if called in setEnv
            # note: this will overwrite existing entries
            items.update(values)

    misc_options.add_argument(
        "--setEnv",
        "-e",
        metavar="NAME=VALUE or NAME",
        dest="environment",
        default={},
        type=yaml_safe_load,
        action=ExtendActionDict,
        help="Set an environment variable early on in the worker. If VALUE is null, it will "
        "be looked up in the current environment. Independently of this option, the worker "
        "will try to emulate the leader's environment before running a job, except for "
        "some variables known to vary across systems.  Using this option, a variable can "
        "be injected into the worker process itself before it is started.",
    )
    misc_options.add_argument(
        "--servicePollingInterval",
        dest="servicePollingInterval",
        default=60.0,
        type=float,
        action=make_open_interval_action(0.0),
        metavar="FLOAT",
        help=f"Interval of time service jobs wait between polling for the existence of the "
        f"keep-alive flag.  Default: {60.0}",
    )
    misc_options.add_argument(
        "--forceDockerAppliance",
        dest="forceDockerAppliance",
        type=strtobool,
        default=False,
        metavar="BOOL",
        help="Disables sanity checking the existence of the docker image specified by "
        "TOIL_APPLIANCE_SELF, which Toil uses to provision mesos for autoscaling.",
    )
    misc_options.add_argument(
        "--statusWait",
        dest="statusWait",
        type=int,
        default=3600,
        metavar="INT",
        help="Seconds to wait between reports of running jobs.",
    )
    misc_options.add_argument(
        "--disableProgress",
        dest="disableProgress",
        action="store_true",
        default=False,
        help="Disables the progress bar shown when standard error is a terminal.",
    )
    misc_options.add_argument(
        "--publishWorkflowMetrics",
        dest="publish_workflow_metrics",
        choices=["all", "current", "no"],
        default=None,
        help="Whether to publish workflow metrics reports (including unique workflow "
        "and task run IDs, job names, and version and Toil feature use information) to "
        "Dockstore when a workflow completes. Selecting \"current\" will publish metrics "
        "for the current workflow. Selecting \"all\" will also publish prior workflow "
        "runs from the Toil history database, even if they themselves were run with \"no\". "
        "Note that once published, workflow metrics CANNOT be deleted or un-published; they "
        "will stay published forever!"
    )

    # Debug options
    debug_options = parser.add_argument_group(
        title="Toil debug options",
        description="Debug options for finding problems or helping with testing.",
    )
    debug_options.add_argument(
        "--debugWorker",
        dest="debugWorker",
        default=False,
        action="store_true",
        help="Experimental no forking mode for local debugging.  Specifically, workers "
        "are not forked and stderr/stdout are not redirected to the log.",
    )
    debug_options.add_argument(
        "--disableWorkerOutputCapture",
        dest="disableWorkerOutputCapture",
        default=False,
        action="store_true",
        help="Let worker output go to worker's standard out/error instead of per-job logs.",
    )
    debug_options.add_argument(
        "--badWorker",
        dest="badWorker",
        default=0.0,
        type=float,
        action=make_closed_interval_action(0.0, 1.0),
        metavar="FLOAT",
        help=f"For testing purposes randomly kill --badWorker proportion of jobs using "
        f"SIGKILL.  default={0.0}",
    )
    debug_options.add_argument(
        "--badWorkerFailInterval",
        dest="badWorkerFailInterval",
        default=0.01,
        type=float,
        action=make_open_interval_action(0.0),
        metavar="FLOAT",  # might be cyclical?
        help=f"When killing the job pick uniformly within the interval from 0.0 to "
        f"--badWorkerFailInterval seconds after the worker starts.  "
        f"default={0.01}",
    )

    # All deprecated options:

    # These are deprecated in favor of a simpler option
    #   ex: noLinkImports and linkImports can be simplified into a single link_imports argument
    link_imports.add_argument(
        "--noLinkImports", dest="linkImports", action="store_false", help=SUPPRESS
    )
    link_imports.add_argument(
        "--linkImports", dest="linkImports", action="store_true", help=SUPPRESS
    )
    link_imports.set_defaults(linkImports=None)

    move_exports.add_argument(
        "--moveExports", dest="moveExports", action="store_true", help=SUPPRESS
    )
    move_exports.add_argument(
        "--noMoveExports", dest="moveExports", action="store_false", help=SUPPRESS
    )
    link_imports.set_defaults(moveExports=None)

    # dest is set to enableCaching to not conflict with the current --caching destination
    caching.add_argument(
        "--disableCaching", dest="enableCaching", action="store_false", help=SUPPRESS
    )
    caching.set_defaults(enableCaching=None)



# Copyright (C) 2024 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains tools for tracking history.
"""

from contextlib import contextmanager
import logging
import os
import sqlite3
import sys
import threading
import time
import uuid
from dataclasses import dataclass
from typing import Any, Iterable, Iterator, Optional, TypeVar, Callable

from toil.lib.conversions import strtobool
from toil.lib.io import get_toil_home
from toil.lib.retry import ErrorCondition, retry

logger = logging.getLogger(__name__)

class HistoryDatabaseSchemaTooNewError(RuntimeError):
    """
    Raised when we would write to the history database, but its schema is too
    new for us to understand.
    """
    pass

@dataclass
class WorkflowSummary:
    """
    Data class holding summary information for a workflow.

    Represents all the attempts to execute one run of a workflow.
    """
    id: str
    name: Optional[str]
    job_store: str
    total_attempts: int
    total_job_attempts: int
    succeeded: bool
    start_time: Optional[float]
    """
    Time when the first workflow attempt started, in seconds since epoch.

    None if there are no attempts recorded.
    """
    runtime: Optional[float]
    """
    Time from the first workflow attempt's start to the last one's end, in seconds.

    None if there are no attempts recorded.
    """
    trs_spec: Optional[str]

@dataclass
class WorkflowAttemptSummary:
    """
    Data class holding summary information for a workflow attempt.

    Helpfully includes the workflow metadata for Dockstore.
    """
    workflow_id: str
    attempt_number: int
    succeeded: bool
    start_time: float
    runtime: float
    submitted_to_dockstore: bool
    batch_system: Optional[str]
    caching: Optional[bool]
    toil_version: Optional[str]
    python_version: Optional[str]
    platform_system: Optional[str]
    platform_machine: Optional[str]
    workflow_job_store: str
    workflow_trs_spec: Optional[str]

@dataclass
class JobAttemptSummary:
    """
    Data class holding summary information for a job attempt within a known
    workflow attempt.
    """
    id: str
    job_name: str
    succeeded: bool
    start_time: float
    runtime: float
    submitted_to_dockstore: bool
    cores: Optional[float]
    cpu_seconds: Optional[float]
    memory_bytes: Optional[int]
    disk_bytes: Optional[int]


RT = TypeVar("RT")

def db_retry(function: Callable[..., RT]) -> Callable[..., RT]:
    """
    Decorate a function with the appropriate retries for accessing the database.
    """
    return retry(
        infinite_retries=True,
        errors=[
            ErrorCondition(
                error=sqlite3.OperationalError, error_message_must_include="is locked"
            )
        ],
    )(function)

class HistoryManager:
    """
    Class responsible for managing the history of Toil runs.
    """

    @classmethod
    def enabled(cls) -> bool:
        """
        Return True if history should be read from and written to the database.

        If False, no access at all shoulf be made to the database.
        """
        return strtobool(os.environ.get("TOIL_HISTORY", 'True')) 

    @classmethod
    def enabled_job(cls) -> bool:
        """
        Return True if job history should be read from and written to the database.

        Always returns False if enabled() returns False.
        """
        # TODO: When Dockstore can take job metrics alongside whole-workflow
        # metrics, and we've tested to make sure history recording doesn't slow
        # down our leader job processing rate, turn on actual job history logging.
        return cls.enabled() and strtobool(os.environ.get("TOIL_JOB_HISTORY", 'False')) 

    # For testing, we can move the database path for the class.
    database_path_override: Optional[str] = None

    @classmethod
    def database_path(cls) -> str:
        """
        Get the path at which the database we store history in lives.
        """
        if cls.database_path_override is not None:
            # Under test, we can use a temporary path.
            return cls.database_path_override

        return os.path.join(get_toil_home(), "history.sqlite")

    @classmethod
    def connection(cls) -> sqlite3.Connection:
        """
        Connect to the history database.

        Caller must not actually use the connection without using
        ensure_tables() to protect reads and updates.

        Must be called from inside a top-level method marked @db_retry.

        The connection will be in DEFERRED isolation_level, with autocommit off
        on Python versions that support it. In order to run any commands
        outside of a transaction use the no_transaction context manager.
        """

        if not cls.enabled():
            # Make sure we're not missing an enabled check along any codepath
            # that wants to access the database.
            raise RuntimeError("Attempting to connect to database when HistoryManager is disabled!")

        if not os.path.exists(cls.database_path()):
            # Make the database and protect it from snoopers and busybodies
            con = sqlite3.connect(cls.database_path())
            del con
            os.chmod(cls.database_path(), 0o600)

        con = sqlite3.connect(
            cls.database_path(),
            isolation_level="DEFERRED"
        )

        with cls.no_transaction(con):
            # Turn on foreign keys.
            # This has to be outside any transaction.
            # See <https://stackoverflow.com/q/78898176>
            con.execute("PRAGMA foreign_keys = ON")
        # This has the side effect of definitely leaving autocommit off, which
        # is what we want as the base state.

        # Set up the connection to use the Row class so that we can look up row values by column name and not just order.
        con.row_factory = sqlite3.Row
        
        return con

    @classmethod
    @contextmanager
    def no_transaction(cls, con: sqlite3.Connection) -> Iterator[None]:
        """
        Temporarily disable the constant active transaction on the database
        connection, on Python versions where it exists.

        Commits the current transaction.
        """
        
        con.commit()
        if hasattr(con, 'autocommit'):
            con.autocommit = True
        yield
        if hasattr(con, 'autocommit'):
            con.autocommit = False

    @classmethod
    def ensure_tables(cls, con: sqlite3.Connection, cur: sqlite3.Cursor) -> None:
        """
        Ensure that tables exist in the database and the schema is migrated to the current version.

        Leaves the cursor in a transaction where the schema version is known to be correct.

        Must be called from inside a top-level methodf marked @db_retry.

        :raises HistoryDatabaseSchemaTooNewError: If the schema is newer than the current version.
        """

        # Python already puts us in a transaction.

        # TODO: Do a try-and-fall-back to avoid sending the table schema for
        # this every time we do anything.
        cur.execute("""
            CREATE TABLE IF NOT EXISTS migrations (
                version INT NOT NULL PRIMARY KEY,
                description TEXT
            )
        """)
        db_version = next(cur.execute("SELECT MAX(version) FROM migrations"))[0]
        if db_version is None:
            db_version = -1

        # This holds pairs of description and command lists.
        # To make a schema change, ADD A NEW PAIR AT THE END, and include
        # statements to adjust existing data.
        migrations = [
            (
                "Make initial tables",
                [
                    """
                    CREATE TABLE workflows (
                        id TEXT NOT NULL PRIMARY KEY,
                        job_store TEXT NOT NULL,
                        creation_time REAL NOT NULL,
                        name TEXT,
                        trs_spec TEXT
                    )
                    """,
                    """
                    CREATE INDEX idx_workflows_by_creation_time
                    ON workflows (creation_time)
                    """,
                    # There's no reference constraint from the job attempts to
                    # the workflow attempts because the jobs for a workflow
                    # attempt need to go in before the attempt is known to be
                    # finished or failed/before the attempt is submittable to
                    # Dockstore.
                    #
                    # TODO: Should we force workflow attempts to be reported on
                    # start so that we can have the jobs key-reference them?
                    # And so that we always have a start time for the workflow
                    # as a whole?
                    """
                    CREATE TABLE job_attempts (
                        id TEXT NOT NULL PRIMARY KEY,
                        workflow_id TEXT NOT NULL,
                        workflow_attempt_number INT NOT NULL,
                        job_name TEXT NOT NULL,
                        succeeded INTEGER NOT NULL,
                        start_time REAL NOT NULL,
                        runtime REAL NOT NULL,
                        cores REAL,
                        cpu_seconds REAL,
                        memory_bytes INTEGER,
                        disk_bytes INTEGER,
                        submitted_to_dockstore INTEGER NOT NULL DEFAULT FALSE,
                        FOREIGN KEY(workflow_id) REFERENCES workflows(id)
                    )
                    """,
                    """
                    CREATE INDEX idx_job_attempts_by_workflow_attempt
                    ON job_attempts (workflow_id, workflow_attempt_number)
                    """,
                    """
                    CREATE TABLE workflow_attempts (
                        workflow_id TEXT NOT NULL,
                        attempt_number INTEGER NOT NULL,
                        succeeded INTEGER NOT NULL,
                        start_time REAL NOT NULL,
                        runtime REAL NOT NULL,
                        batch_system TEXT,
                        caching INTEGER,
                        toil_version TEXT,
                        python_version TEXT,
                        platform_system TEXT,
                        platform_machine TEXT,
                        submitted_to_dockstore INTEGER NOT NULL DEFAULT FALSE,
                        PRIMARY KEY(workflow_id,attempt_number),
                        FOREIGN KEY(workflow_id) REFERENCES workflows(id)
                    )
                    """
                ],
            ),
        ]

        if db_version + 1 > len(migrations):
            raise HistoryDatabaseSchemaTooNewError(f"History database version is {db_version}, but known migrations only go up to {len(migrations) - 1}")

        for migration_number in range(db_version + 1, len(migrations)):
            for statement_number, statement in enumerate(migrations[migration_number][1]):
                # Run all the migration commands.
                # We don't use executescript() because (on old Pythons?) it
                # commits the current transactrion first.
                try:
                    cur.execute(statement)
                except sqlite3.OperationalError:
                    logger.exception("Could not execute migration %s statement %s: %s", migration_number, statement_number, statement)
                    raise
            cur.execute("INSERT INTO migrations VALUES (?, ?)", (migration_number, migrations[migration_number][0]))

        # If we did have to migrate, leave everything else we do as part of the migration transaction.

    ##
    # Recording Methods
    ##

    @classmethod
    @db_retry
    def record_workflow_creation(cls, workflow_id: str, job_store_spec: str) -> None:
        """
        Record that a workflow is being run.

        Takes the Toil config's workflow ID and the location of the job store.

        Should only be called on the *first* attempt on a job store, not on a
        restart.

        A workflow may have multiple attempts to run it, some of which succeed
        and others of which fail. Probably only the last one should succeed.

        :param job_store_spec: The job store specifier for the workflow. Should
            be canonical and always start with the type and a colon. If the
            job store is later moved by the user, the location will not be
            updated.
        """

        if not cls.enabled():
            return

        logger.info("Recording workflow creation of %s in %s", workflow_id, job_store_spec)

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)
            cur.execute("INSERT INTO workflows VALUES (?, ?, ?, NULL, NULL)", (workflow_id, job_store_spec, time.time()))
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

        # If we raise out of here the connection goes away and the transaction rolls back.


    @classmethod
    @db_retry
    def record_workflow_metadata(cls, workflow_id: str, workflow_name: str, trs_spec: Optional[str] = None) -> None:
        """
        Associate a name and optionally a TRS ID and version with a workflow run.
        """

        # TODO: Make name of this function less general?

        if not cls.enabled():
            return

        logger.info("Workflow %s is a run of %s", workflow_id, workflow_name)
        if trs_spec:
            logger.info("Workflow %s has TRS ID and version %s", workflow_id, trs_spec)

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)
            cur.execute("UPDATE workflows SET name = ? WHERE id = ?", (workflow_name, workflow_id))
            if trs_spec is not None:
                cur.execute("UPDATE workflows SET trs_spec = ? WHERE id = ?", (trs_spec, workflow_id))
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

    @classmethod
    @db_retry
    def record_job_attempt(
            cls,
            workflow_id: str,
            workflow_attempt_number: int,
            job_name: str,
            succeeded: bool,
            start_time: float,
            runtime: float,
            cores: Optional[float] = None,
            cpu_seconds: Optional[float] = None,
            memory_bytes: Optional[int] = None,
            disk_bytes: Optional[int] = None
        ) -> None:
        """
        Record that a job ran in a workflow.

        Doesn't expect the provided information to uniquely identify the job
        attempt; assigns the job attempt its own unique ID.

        Thread safe.

        :param job_name: A human-readable name for the job. Not expected to be
            a job store ID or to necessarily uniquely identify the job within
            the workflow.
        :param start_time: Job execution start time ins econds since epoch.
        :param runtime: Job execution duration in seconds.
        :param cores: Number of CPU cores the job was scheduled on.
        :param cpu_seconds: CPU core-seconds actually consumed.
        :param memory_bytes: Peak observed job memory usage.
        :param disk_bytes: Observed job disk usage.
        """

        if not cls.enabled_job():
            return

        logger.debug("Workflow %s ran job %s", workflow_id, job_name)

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)
            cur.execute(
                """
                INSERT INTO job_attempts(
                    id,
                    workflow_id,
                    workflow_attempt_number,
                    job_name,
                    succeeded,
                    start_time,
                    runtime,
                    cores,
                    cpu_seconds,
                    memory_bytes,
                    disk_bytes
                )
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                """,
                (
                    str(uuid.uuid4()),
                    workflow_id,
                    workflow_attempt_number,
                    job_name,
                    1 if succeeded else 0,
                    start_time,
                    runtime,
                    cores,
                    cpu_seconds,
                    memory_bytes,
                    disk_bytes,
                )
            )
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

    @classmethod
    @db_retry
    def record_workflow_attempt(
        cls,
        workflow_id: str,
        workflow_attempt_number: int,
        succeeded: bool,
        start_time: float,
        runtime: float,
        batch_system: Optional[str] = None,
        caching: Optional[bool] = None,
        toil_version: Optional[str] = None,
        python_version: Optional[str] = None,
        platform_system: Optional[str] = None,
        platform_machine: Optional[str] = None
    ) -> None:
        """
        Record a workflow attempt (start or restart) having finished or failed.

        :param batch_system: The Python type name of the batch system implementation used.
        :param caching: Whether Toil filestore-level caching was used.
        :param toil_version: Version of Toil used to run the workflow.
        :param python_version: Version of Python used to run the workflow.
        :param platform_system: OS ("Darwin", "Linux", etc.) used to run the workflow.
        :param platform_machine: CPU type ("AMD64", etc.) used to run the workflow leader.
        """

        if not cls.enabled():
            return

        logger.info("Workflow %s stopped. Success: %s", workflow_id, succeeded)

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)
            cur.execute(
                """
                INSERT INTO workflow_attempts(
                    workflow_id,
                    attempt_number,
                    succeeded,
                    start_time,
                    runtime,
                    batch_system,
                    caching,
                    toil_version,
                    python_version,
                    platform_system,
                    platform_machine
                )
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
                """,
                (
                    workflow_id,
                    workflow_attempt_number,
                    1 if succeeded else 0,
                    start_time,
                    runtime,
                    batch_system,
                    caching,
                    toil_version,
                    python_version,
                    platform_system,
                    platform_machine
                )
            )
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()


    ##
    # Read methods
    ##

    # We would implement a bunch of iterators and allow follow-up queries, but
    # then we'd have to figure out how to make sure we use one connection and
    # cursor and not block ourselves with the database transaction locks.
    #
    # So instead we always fetch all the information asked for and close out
    # the read transaction before returning.
    #
    # This means the caller has to worry about a workflow vanishing or changing
    # between when it was shown to them and when they ask follow-up questions,
    # but it also means we can't deadlock.

    @classmethod
    @db_retry
    def summarize_workflows(cls) -> list[WorkflowSummary]:
        """
        List all known workflows and their summary statistics.
        """

        if not cls.enabled():
            return []

        workflows = []

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)
            cur.execute(
                """
                SELECT
                    workflows.id AS id,
                    workflows.name AS name,
                    workflows.job_store AS job_store,
                    (SELECT count(*) FROM workflow_attempts WHERE workflow_id = workflows.id) AS total_attempts,
                    (SELECT count(*) FROM job_attempts WHERE workflow_id = workflows.id) AS total_job_attempts,
                    (SELECT min(count(*), 1) FROM workflow_attempts WHERE workflow_id = workflows.id AND succeeded = TRUE) AS succeeded,
                    (SELECT min(start_time) FROM workflow_attempts WHERE workflow_id = workflows.id) AS start_time,
                    (SELECT max(start_time + runtime) FROM workflow_attempts WHERE workflow_id = workflows.id) AS end_time,
                    workflows.trs_spec AS trs_spec
                FROM workflows
                ORDER BY start_time DESC
                """
            )
            for row in cur:
                workflows.append(
                    WorkflowSummary(
                        id=row["id"],
                        name=row["name"],
                        job_store=row["job_store"],
                        total_attempts=row["total_attempts"],
                        total_job_attempts=row["total_job_attempts"],
                        succeeded=(row["succeeded"] == 1),
                        start_time=row["start_time"],
                        runtime=(row["end_time"] - row["start_time"]) if row["start_time"] is not None and row["end_time"] is not None else None,
                        trs_spec=row["trs_spec"]
                    )
                )
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

        return workflows

    @classmethod
    @db_retry
    def get_submittable_workflow_attempts(cls, limit: int = sys.maxsize) -> list[WorkflowAttemptSummary]:
        """
        List all workflow attempts not yet submitted to Dockstore.

        :param limit: Get no more than this many.
        """

        if not cls.enabled():
            return []

        attempts = []

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)
            cur.execute(
                """
                SELECT
                    workflow_attempts.workflow_id AS workflow_id,
                    workflow_attempts.attempt_number AS attempt_number,
                    workflow_attempts.succeeded AS succeeded,
                    workflow_attempts.start_time AS start_time,
                    workflow_attempts.runtime AS runtime,
                    workflow_attempts.batch_system AS batch_system,
                    workflow_attempts.caching AS caching,
                    workflow_attempts.toil_version AS toil_version,
                    workflow_attempts.python_version AS python_version,
                    workflow_attempts.platform_system AS platform_system,
                    workflow_attempts.platform_machine AS platform_machine,
                    workflow_attempts.submitted_to_dockstore AS submitted_to_dockstore,
                    workflows.job_store AS workflow_job_store,
                    workflows.trs_spec AS workflow_trs_spec
                FROM workflow_attempts
                    JOIN workflows ON workflow_attempts.workflow_id = workflows.id
                WHERE workflow_attempts.submitted_to_dockstore = FALSE
                    AND workflows.trs_spec IS NOT NULL
                ORDER BY start_time DESC
                LIMIT ?
                """,
                (limit,)
            )
            for row in cur:
                attempts.append(
                    WorkflowAttemptSummary(
                        workflow_id=row["workflow_id"],
                        attempt_number=row["attempt_number"],
                        succeeded=(row["succeeded"] == 1),
                        start_time=row["start_time"],
                        runtime=row["runtime"],
                        batch_system=row["batch_system"],
                        caching=(row["caching"] == 1),
                        toil_version=row["toil_version"],
                        python_version=row["python_version"],
                        platform_system=row["platform_system"],
                        platform_machine=row["platform_machine"],
                        submitted_to_dockstore=(row["submitted_to_dockstore"] == 1),
                        workflow_job_store=row["workflow_job_store"],
                        workflow_trs_spec=row["workflow_trs_spec"]
                    )
                )
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

        return attempts

    @classmethod
    @db_retry
    def get_workflow_attempts_with_submittable_job_attempts(cls, limit: int = sys.maxsize) -> list[WorkflowAttemptSummary]:
        """
        Get all workflow attempts that have job attempts not yet submitted to
        Dockstore.

        The workflow attempts themselves will have finished and been recorded,
        and have TRS IDs.

        :param limit: Get no more than this many.
        """

        if not cls.enabled_job():
            return []

        attempts = []

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)
            cur.execute(
                """
                SELECT
                    workflow_attempts.workflow_id AS workflow_id,
                    workflow_attempts.attempt_number AS attempt_number,
                    workflow_attempts.succeeded AS succeeded,
                    workflow_attempts.start_time AS start_time,
                    workflow_attempts.runtime AS runtime,
                    workflow_attempts.batch_system AS batch_system,
                    workflow_attempts.caching AS caching,
                    workflow_attempts.toil_version AS toil_version,
                    workflow_attempts.python_version AS python_version,
                    workflow_attempts.platform_system AS platform_system,
                    workflow_attempts.platform_machine AS platform_machine,
                    workflow_attempts.submitted_to_dockstore AS submitted_to_dockstore,
                    workflows.job_store AS workflow_job_store,
                    workflows.trs_spec AS workflow_trs_spec
                FROM (
                    SELECT DISTINCT
                        workflow_id, workflow_attempt_number
                    FROM job_attempts
                    WHERE job_attempts.submitted_to_dockstore = FALSE
                ) AS found_job_attempts
                    JOIN workflows ON found_job_attempts.workflow_id = workflows.id
                    JOIN workflow_attempts ON
                        found_job_attempts.workflow_id = workflow_attempts.workflow_id
                        AND found_job_attempts.workflow_attempt_number = workflow_attempts.attempt_number
                WHERE workflows.trs_spec IS NOT NULL
                LIMIT ?
                """,
                (limit,)
            )
            for row in cur:
                # TODO: Unify row to data class conversion
                attempts.append(
                    WorkflowAttemptSummary(
                        workflow_id=row["workflow_id"],
                        attempt_number=row["attempt_number"],
                        succeeded=(row["succeeded"] == 1),
                        start_time=row["start_time"],
                        runtime=row["runtime"],
                        batch_system=row["batch_system"],
                        caching=(row["caching"] == 1),
                        toil_version=row["toil_version"],
                        python_version=row["python_version"],
                        platform_system=row["platform_system"],
                        platform_machine=row["platform_machine"],
                        submitted_to_dockstore=(row["submitted_to_dockstore"] == 1),
                        workflow_job_store=row["workflow_job_store"],
                        workflow_trs_spec=row["workflow_trs_spec"]
                    )
                )
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

        return attempts

    @classmethod
    @db_retry
    def get_workflow_attempt(cls, workflow_id: str, attempt_number: int) -> Optional[WorkflowAttemptSummary]:
        """
        Get a single (not necessarily unsubmitted, not necessarily TRS-ID-having) workflow attempt summary, if present.
        """

        # TODO: Consolidate with the other 2 ways to query workflow attempts!

        if not cls.enabled():
            return None

        attempts = []

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)
            cur.execute(
                """
                SELECT
                    workflow_attempts.workflow_id AS workflow_id,
                    workflow_attempts.attempt_number AS attempt_number,
                    workflow_attempts.succeeded AS succeeded,
                    workflow_attempts.start_time AS start_time,
                    workflow_attempts.runtime AS runtime,
                    workflow_attempts.batch_system AS batch_system,
                    workflow_attempts.caching AS caching,
                    workflow_attempts.toil_version AS toil_version,
                    workflow_attempts.python_version AS python_version,
                    workflow_attempts.platform_system AS platform_system,
                    workflow_attempts.platform_machine AS platform_machine,
                    workflow_attempts.submitted_to_dockstore AS submitted_to_dockstore,
                    workflows.job_store AS workflow_job_store,
                    workflows.trs_spec AS workflow_trs_spec
                FROM workflow_attempts
                    JOIN workflows ON workflow_attempts.workflow_id = workflows.id
                WHERE workflow_id = ?
                    AND attempt_number = ?
                ORDER BY start_time DESC
                LIMIT 1
                """,
                (workflow_id, attempt_number)
            )
            for row in cur:
                attempts.append(
                    WorkflowAttemptSummary(
                        workflow_id=row["workflow_id"],
                        attempt_number=row["attempt_number"],
                        succeeded=(row["succeeded"] == 1),
                        start_time=row["start_time"],
                        runtime=row["runtime"],
                        batch_system=row["batch_system"],
                        caching=(row["caching"] == 1),
                        toil_version=row["toil_version"],
                        python_version=row["python_version"],
                        platform_system=row["platform_system"],
                        platform_machine=row["platform_machine"],
                        submitted_to_dockstore=(row["submitted_to_dockstore"] == 1),
                        workflow_job_store=row["workflow_job_store"],
                        workflow_trs_spec=row["workflow_trs_spec"]
                    )
                )
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

        if len(attempts) == 0:
            # Not found
            return None
        else:
            return attempts[0]

    @classmethod
    @db_retry
    def get_unsubmitted_job_attempts(cls, workflow_id: str, attempt_number: int) -> list[JobAttemptSummary]:
        """
        List all job attempts in the given workflow attempt not yet submitted to Dockstore.

        Doesn't check to make sure the workflow has a TRS ID.
        """

        if not cls.enabled_job():
            return []

        attempts = []

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)
            cur.execute(
                """
                SELECT
                    id,
                    job_name,
                    succeeded,
                    start_time,
                    runtime,
                    cores,
                    cpu_seconds,
                    memory_bytes,
                    disk_bytes,
                    submitted_to_dockstore
                FROM job_attempts
                WHERE workflow_id = ?
                AND workflow_attempt_number = ?
                AND submitted_to_dockstore = FALSE
                ORDER BY start_time DESC
                """,
                (workflow_id, attempt_number)
            )
            for row in cur:
                attempts.append(
                    JobAttemptSummary(
                        id=row["id"],
                        job_name=row["job_name"],
                        succeeded=(row["succeeded"] == 1),
                        start_time=row["start_time"],
                        runtime=row["runtime"],
                        cores=row["cores"],
                        cpu_seconds=row["cpu_seconds"],
                        memory_bytes=row["memory_bytes"],
                        disk_bytes=row["disk_bytes"],
                        submitted_to_dockstore=(row["submitted_to_dockstore"] == 1)
                    )
                )
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

        return attempts

    ###
    # Submission marking methods
    ###

    @classmethod
    @db_retry
    def mark_workflow_attempt_submitted(cls, workflow_id: str, attempt_number: int) -> None:
        """
        Mark a workflow attempt as having been successfully submitted to Dockstore.

        Does not mark the workflow attempt's job attempts as submitted.
        """

        if not cls.enabled():
            return

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)
            cur.execute(
                "UPDATE workflow_attempts SET submitted_to_dockstore = TRUE WHERE workflow_id = ? AND attempt_number = ?",
                (workflow_id, attempt_number)
            )
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

    @classmethod
    @db_retry
    def mark_job_attempts_submitted(cls, job_attempt_ids: list[str]) -> None:
        """
        Mark a collection of job attempts as submitted to Dockstore in a single transaction.
        """

        if not cls.enabled_job():
            return

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)
            for job_attempt_id in job_attempt_ids:
                # Do all the marking in one transaction
                cur.execute(
                    "UPDATE job_attempts SET submitted_to_dockstore = TRUE WHERE id = ?",
                    (job_attempt_id,)
                )
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

    @classmethod
    @db_retry
    def count_workflows(cls) -> int:
        """
        Count workflows in the database.
        """
        
        if not cls.enabled():
            return 0

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)

            cur.execute("SELECT count(*) FROM workflows")

            count = cur.fetchone()[0]
            assert isinstance(count, int)
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

        return count

    @classmethod
    @db_retry
    def count_workflow_attempts(cls) -> int:
        """
        Count workflow attempts in the database.
        """

        if not cls.enabled():
            return 0

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)

            cur.execute("SELECT count(*) FROM workflow_attempts")

            count = cur.fetchone()[0]
            assert isinstance(count, int)
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

        return count

    @classmethod
    @db_retry
    def count_job_attempts(cls) -> int:
        """
        Count job attempts in the database.
        """

        if not cls.enabled_job():
            return 0

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)

            cur.execute("SELECT count(*) FROM job_attempts")

            count = cur.fetchone()[0]
            assert isinstance(count, int)
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

        return count

    @classmethod
    @db_retry
    def get_fully_submitted_workflow_ids(cls, limit: int = sys.maxsize) -> list[str]:
        """
        Get workflows that have a successful attempt and no unsubmitted attempts or job attempts.
        """

        if not cls.enabled():
            return []

        ids = []

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)

            cur.execute(
                """
                SELECT
                    workflows.id
                FROM workflows
                WHERE
                    (
                        SELECT
                            count(*)
                        FROM workflow_attempts
                        WHERE workflow_id = workflows.id
                            AND succeeded = TRUE
                            AND submitted_to_dockstore = TRUE
                        LIMIT 1
                    ) = 1
                    AND (
                        SELECT
                            count(*)
                        FROM workflow_attempts
                        WHERE workflow_id = workflows.id
                            AND submitted_to_dockstore = FALSE
                        LIMIT 1
                    ) = 0
                    AND (
                        SELECT
                            count(*)
                        FROM job_attempts
                        WHERE workflow_id = workflows.id
                            AND submitted_to_dockstore = FALSE
                        LIMIT 1
                    ) = 0
                LIMIT ?
                """,
                (limit,)
            )
            for row in cur:
                ids.append(row["id"])
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

        return ids

    @classmethod
    @db_retry
    def get_oldest_workflow_ids(cls, limit: int = sys.maxsize) -> list[str]:
        """
        Get workflows that are old.
        """

        if not cls.enabled():
            return []

        ids = []

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)

            # We could use a complicated query to bump workflows down the list
            # if they have been updated by having attempts or job attempts. But
            # that would mean we'd need to do a lot of querying and live
            # sorting, whereas using just the creation time lets us use an
            # index and a limit efficiently.

            cur.execute(
                """
                SELECT
                    id,
                    creation_time
                FROM workflows
                ORDER BY creation_time ASC
                LIMIT ?
                """,
                (limit,)
            )
            for row in cur:
                ids.append(row["id"])
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

        return ids

    @classmethod
    @db_retry
    def delete_workflow(cls, workflow_id: str) -> None:
        """
        Delete a workflow and all its attempts and job attempts.

        Succeeds if the workflow does not exist.
        """

        if not cls.enabled():
            return

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)

            cur.execute("DELETE FROM job_attempts WHERE workflow_id = ?", (workflow_id,))
            cur.execute("DELETE FROM workflow_attempts WHERE workflow_id = ?", (workflow_id,))
            cur.execute("DELETE FROM workflows WHERE id = ?", (workflow_id,))
        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

    @classmethod
    @db_retry
    def get_database_byte_size(cls) -> int:
        """
        Get the total number of bytes used by the database.
        """

        if not cls.enabled():
            return 0

        con = cls.connection()
        cur = con.cursor()
        try:
            cls.ensure_tables(con, cur)

            cur.execute("PRAGMA page_size")
            page_size = cur.fetchone()[0]
            assert isinstance(page_size, int)

            cur.execute("PRAGMA page_count")
            page_count = cur.fetchone()[0]
            assert isinstance(page_count, int)

        except:
            con.rollback()
            con.close()
            raise
        else:
            con.commit()
            con.close()

        return page_size * page_count

    @classmethod
    @db_retry
    def compact_database(cls) -> None:
        """
        Shrink the database to remove unused space.
        """

        if not cls.enabled():
            return
        
        con = cls.connection()
        cur = con.cursor()

        # Don't bother making tables; we don't need them for this and they need
        # a transaction.
        
        with cls.no_transaction(con):
            # Do the vacuum outside any transaction, and rely on it to
            # synchronize appropriately internally.
            cur.execute("VACUUM")

        con.close()

    @classmethod
    def enforce_byte_size_limit(cls, limit: int = 100 * 1024 * 1024) -> None:
        """
        Shrink the database until it is smaller than the given limit, or until
        it is empty, by throwing away workflows.

        Throws data away in a sensible order, least important to most
        important.
        """

        if not cls.enabled():
            return

        db_size = cls.get_database_byte_size()

        if db_size < limit:
            # Nothing to do!
            return

        while db_size > limit:
            # Look for some things we submitted already
            target_workflows = cls.get_fully_submitted_workflow_ids(limit=100)
            if len(target_workflows) == 0:
                # If there aren't any, do oldest workflows a few at a time
                # We need to balance the O(n^2)
                # delete-and-copy-the-whole-db-to-vacuum loop with not wanting
                # to delete too many workflows we could keep.
                target_workflows = cls.get_oldest_workflow_ids(limit=10)
            if len(target_workflows) == 0:
                # There are no more workflows to delete.
                break

            for workflow_id in target_workflows:
                # Delete all the workflows we don't want.
                cls.delete_workflow(workflow_id)

            # Shrink the DB
            cls.compact_database()
            # Re-check the size
            db_size = cls.get_database_byte_size()




    @classmethod
    def database_dump_lines(cls) -> Iterable[str]:
        """
        Yield lines from the database dump.

        For debugging tests.
        """

        if not cls.enabled():
            return []

        return cls.connection().iterdump()







# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

import pytest

from toil.provisioners import parse_node_types
from toil.test import ToilTest

log = logging.getLogger(__name__)


class ProvisionerTest(ToilTest):

    def test_node_type_parsing(self) -> None:
        assert parse_node_types(None) == []
        assert parse_node_types("") == []
        assert parse_node_types("red beans") == [({"red beans"}, None)]
        assert parse_node_types("red beans,rice") == [
            ({"red beans"}, None),
            ({"rice"}, None),
        ]
        assert parse_node_types("red beans/black beans,rice") == [
            ({"red beans", "black beans"}, None),
            ({"rice"}, None),
        ]
        assert parse_node_types("frankfurters:0.05") == [({"frankfurters"}, 0.05)]
        assert parse_node_types(
            "red beans/black beans:999,rice,red beans/black beans"
        ) == [
            ({"red beans", "black beans"}, 999),
            ({"rice"}, None),
            ({"red beans", "black beans"}, None),
        ]
        with pytest.raises(ValueError):
            parse_node_types("your thoughts:penny")
        with pytest.raises(ValueError) as err:
            parse_node_types(",,,")
        assert "empty" in str(err.value)
        with pytest.raises(ValueError):
            parse_node_types("now hear this:")
        with pytest.raises(ValueError) as err:
            parse_node_types("miles I will walk:500:500")
        assert "multiple" in str(err.value)
        with pytest.raises(ValueError):
            parse_node_types("red beans:500/black beans:500,rice")



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nacl
from nacl.secret import SecretBox

# 16-byte MAC plus a nonce is added to every message.
overhead = 16 + SecretBox.NONCE_SIZE


def encrypt(message: bytes, keyPath: str) -> bytes:
    """
    Encrypts a message given a path to a local file containing a key.

    :param message: The message to be encrypted.
    :param keyPath: A path to a file containing a 256-bit key (and nothing else).
    :type message: bytes
    :type keyPath: str
    :rtype: bytes

    A constant overhead is added to every encrypted message (for the nonce and MAC).
    >>> import tempfile
    >>> k = tempfile.mktemp()
    >>> with open(k, 'wb') as f:
    ...     _ = f.write(nacl.utils.random(SecretBox.KEY_SIZE))
    >>> message = 'test'.encode('utf-8')
    >>> len(encrypt(message, k)) == overhead + len(message)
    True
    >>> import os
    >>> os.remove(k)
    """
    with open(keyPath, "rb") as f:
        key = f.read()
    if len(key) != SecretBox.KEY_SIZE:
        raise ValueError(
            "Key is %d bytes, but must be exactly %d bytes"
            % (len(key), SecretBox.KEY_SIZE)
        )
    sb = SecretBox(key)
    # We generate the nonce using secure random bits. For long enough
    # nonce size, the chance of a random nonce collision becomes
    # *much* smaller than the chance of a subtle coding error causing
    # a nonce reuse. Currently the nonce size is 192 bits--the chance
    # of a collision is astronomically low. (This approach is
    # recommended in the libsodium documentation.)
    nonce = nacl.utils.random(SecretBox.NONCE_SIZE)
    if len(nonce) != SecretBox.NONCE_SIZE:
        raise RuntimeError("Generated nonce is the wrong size.")
    return bytes(sb.encrypt(message, nonce))


def decrypt(ciphertext: bytes, keyPath: str) -> bytes:
    """
    Decrypts a given message that was encrypted with the encrypt() method.

    :param ciphertext: The encrypted message (as a string).
    :param keyPath: A path to a file containing a 256-bit key (and nothing else).
    :type ciphertext: bytes
    :type keyPath: str
    :rtype: bytes

    Raises an error if ciphertext was modified
    >>> import tempfile
    >>> k = tempfile.mktemp()
    >>> with open(k, 'wb') as f:
    ...     _ = f.write(nacl.utils.random(SecretBox.KEY_SIZE))
    >>> ciphertext = encrypt("testMessage".encode('utf-8'), k)
    >>> ciphertext = b'5' + ciphertext[1:]
    >>> decrypt(ciphertext, k) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    CryptoError: Decryption failed. Ciphertext failed verification

    Otherwise works correctly
    >>> decrypt(encrypt("testMessage".encode('utf-8'), k), k).decode('utf-8') in (u'testMessage', b'testMessage', 'testMessage') # doctest: +ALLOW_UNICODE
    True

    >>> import os
    >>> os.remove(k)
    """
    with open(keyPath, "rb") as f:
        key = f.read()
    if len(key) != SecretBox.KEY_SIZE:
        raise ValueError(
            "Key is %d bytes, but must be exactly %d bytes"
            % (len(key), SecretBox.KEY_SIZE)
        )
    sb = SecretBox(key)
    # The nonce is kept with the message.
    return sb.decrypt(ciphertext)



from typing import Any, Optional, Union

from tes.models import (
    CancelTaskRequest,
    CreateTaskResponse,
    GetTaskRequest,
    ListTasksRequest,
    ListTasksResponse,
    ServiceInfo,
    Task,
    strconv,
)
from tes.utils import TimeoutError, raise_for_status, unmarshal

def process_url(value: str) -> str: ...

class HTTPClient:
    url: str = ...
    timeout: int = ...
    user: Optional[str] = ...
    password: Optional[str] = ...
    token: Optional[str] = ...
    def get_service_info(self) -> ServiceInfo: ...
    def create_task(self, task: Task) -> CreateTaskResponse: ...
    def get_task(self, task_id: str, view: str = ...) -> Task: ...
    def cancel_task(self, task_id: str) -> None: ...
    def list_tasks(
        self,
        view: str = ...,
        page_size: Optional[int] = ...,
        page_token: Optional[str] = ...,
    ) -> ListTasksResponse: ...
    def wait(self, task_id: str, timeout: Union[int, float, None] = ...) -> None: ...
    def __init__(
        self,
        url: str,
        timeout: Optional[int],
        user: Optional[str],
        password: Optional[str],
        token: Optional[str],
    ) -> None: ...



#!/usr/bin/env python3
"""
Runs mypy and ignores files that do not yet have passing type hints.

Does not type check test files (any path including "src/toil/test").
"""
import os
import subprocess
import sys

os.environ['MYPYPATH'] = 'contrib/mypy-stubs'
pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))  # noqa
sys.path.insert(0, pkg_root)  # noqa

from src.toil.lib.resources import glob  # type: ignore


def main():
    all_files_to_check = []
    for d in ['dashboard', 'docker', 'docs', 'src']:
        all_files_to_check += glob(glob_pattern='*.py', directoryname=os.path.join(pkg_root, d))

    # TODO: Remove these paths as typing is added and mypy conflicts are addressed.
    # These are handled as path prefixes.
    ignore_paths = [os.path.abspath(f) for f in [
        'docs/_build',
        'docker/Dockerfile.py',
        'docs/conf.py',
        'docs/vendor/sphinxcontrib/fulltoc.py',
        'docs/vendor/sphinxcontrib/__init__.py',
        'src/toil/job.py',
        'src/toil/leader.py',
        'src/toil/__init__.py',
        'src/toil/deferred.py',
        'src/toil/version.py',
        'src/toil/provisioners/abstractProvisioner.py',
        'src/toil/provisioners/gceProvisioner.py',
        'src/toil/provisioners/__init__.py',
        'src/toil/provisioners/node.py',
        'src/toil/provisioners/aws/boto2Context.py',
        'src/toil/provisioners/aws/__init__.py',
        'src/toil/batchSystems/gridengine.py',
        'src/toil/batchSystems/singleMachine.py',
        'src/toil/batchSystems/torque.py',
        'src/toil/batchSystems/options.py',
        'src/toil/batchSystems/registry.py',
        'src/toil/batchSystems/lsf.py',
        'src/toil/batchSystems/__init__.py',
        'src/toil/batchSystems/abstractGridEngineBatchSystem.py',
        'src/toil/batchSystems/awsBatch.py',
        'src/toil/batchSystems/lsfHelper.py',
        'src/toil/batchSystems/htcondor.py',
        'src/toil/batchSystems/mesos/batchSystem.py',
        'src/toil/batchSystems/mesos/executor.py',
        'src/toil/batchSystems/mesos/conftest.py',
        'src/toil/batchSystems/mesos/__init__.py',
        'src/toil/batchSystems/mesos/test/__init__.py',
        'src/toil/fileStores/cachingFileStore.py',
        'src/toil/jobStores/utils.py',
        'src/toil/jobStores/conftest.py',
        'src/toil/jobStores/fileJobStore.py',
        'src/toil/jobStores/__init__.py',
        'src/toil/jobStores/googleJobStore.py',
        'src/toil/jobStores/aws/utils.py',
        'src/toil/jobStores/aws/jobStore.py',
        'src/toil/jobStores/aws/__init__.py',
        'src/toil/utils/__init__.py',
        'src/toil/lib/throttle.py',
        'src/toil/lib/iterables.py',
        'src/toil/lib/bioio.py',
        'src/toil/lib/ec2.py',
        'src/toil/lib/expando.py',
        'src/toil/lib/exceptions.py',
        'src/toil/lib/__init__.py',
        'src/toil/lib/generatedEC2Lists.py',
        'src/toil/lib/retry.py',
        'src/toil/lib/objects.py',
        'src/toil/lib/encryption/_dummy.py',
        'src/toil/lib/encryption/conftest.py',
        'src/toil/lib/encryption/__init__.py',
        'src/toil/lib/aws/__init__.py',
        'src/toil/server/utils.py',
        'src/toil/test/batchSystems/batch_system_plugin_test.py',
        'src/toil/test/batchSystems/test_gridengine.py',
        'src/toil/test/batchSystems/test_lsf_helper.py',
        'src/toil/test/batchSystems/test_slurm.py',
        'src/toil/test/cactus',
        'src/toil/test/docs',
        'src/toil/test/jobStores',
        'src/toil/test/lib/aws',
        'src/toil/test/lib/test_conversions.py',
        'src/toil/test/lib/test_ec2.py',
        'src/toil/test/lib/test_misc.py',
        'src/toil/test/lib/test_trs.py',
        'src/toil/test/mesos',
        'src/toil/test/options',
        'src/toil/test/provisioners',
        'src/toil/test/server',
        'src/toil/test/sort',
        'src/toil/test/src/realtimeLoggerTest.py',
        'src/toil/test/src/jobDescriptionTest.py',
        'src/toil/test/src/autoDeploymentTest.py',
        'src/toil/test/src/dockerCheckTest.py',
        'src/toil/test/src/fileStoreTest.py',
        'src/toil/test/src/__init__.py',
        'src/toil/test/src/promisesTest.py',
        'src/toil/test/src/checkpointTest.py',
        'src/toil/test/src/jobFileStoreTest.py',
        'src/toil/test/src/userDefinedJobArgTypeTest.py',
        'src/toil/test/src/helloWorldTest.py',
        'src/toil/test/src/toilContextManagerTest.py',
        'src/toil/test/src/environmentTest.py',
        'src/toil/test/src/workerTest.py',
        'src/toil/test/utils/ABCWorkflowDebug',
        'src/toil/utils/toilStats.py'
    ]]

    def ignore(file_path):
        """
        Return True if a file should be ignored.
        """
        for prefix in ignore_paths:
            if file_path.startswith(prefix):
                return True
        return False

    filtered_files_to_check = []
    for file_path in all_files_to_check:
        if not ignore(file_path):
            filtered_files_to_check.append(file_path)
    args = ['mypy', '--color-output', '--show-traceback'] + filtered_files_to_check
    p = subprocess.run(args=args)
    exit(p.returncode)


if __name__ == '__main__':
    main()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Debug tool for running a toil job locally."""
import gc
import logging
import os
import pprint
import sys
from pathlib import Path
from typing import Optional

from toil.common import Toil, parser_with_common_options
from toil.job import FilesDownloadedStoppingPointReached
from toil.jobStores.fileJobStore import FileJobStore
from toil.statsAndLogging import set_logging_from_options
from toil.utils.toilDebugFile import printContentsOfJobStore
from toil.utils.toilStatus import ToilStatus
from toil.worker import workerScript

logger = logging.getLogger(__name__)


def main() -> None:
    parser = parser_with_common_options(
        jobstore_option=True, prog="toil debug-job", default_log_level=logging.DEBUG
    )
    parser.add_argument(
        "job",
        type=str,
        help="The job store id or job name of a job within the provided jobstore",
    )
    parser.add_argument(
        "--printJobInfo",
        action="store_true",
        help="Dump debugging info about the job instead of running it",
    )
    parser.add_argument(
        "--retrieveTaskDirectory",
        dest="retrieve_task_directory",
        type=str,
        default=None,
        help="Download CWL or WDL task inputs to the given directory and stop.",
    )

    options = parser.parse_args()
    set_logging_from_options(options)

    if options.retrieve_task_directory is not None and os.path.exists(
        options.retrieve_task_directory
    ):
        # The logic to duplicate container mounts depends on stuff not already existing.
        logger.error(
            "The directory %s given for --retrieveTaskDirectory already exists. "
            "Stopping to avoid clobbering existing files.",
            options.retrieve_task_directory,
        )
        sys.exit(1)

    jobStore = Toil.resumeJobStore(options.jobStore)
    # Get the config with the workflow ID from the job store
    config = jobStore.config
    # But override its options
    config.setOptions(options)
    config.cleanWorkDir = "never"

    # Find the job

    if jobStore.job_exists(options.job):
        # The user asked for a particular job and it exists
        job_id = options.job
    else:
        # Go search by name and fill in job_id

        # TODO: break out job store scan logic so it doesn't need to re-connect
        # to the job store.
        status = ToilStatus(options.jobStore)
        hits = []
        suggestion = None
        for job in status.jobsToReport:
            if options.job in (job.jobName, job.unitName, job.displayName):
                # Find all the jobs that sort of match
                hits.append(job)
            if suggestion is None and job.remainingTryCount == 0:
                # How about this nice failing job instead?
                suggestion = job
        if len(hits) == 0:
            # No hits
            if suggestion is None:
                logger.critical(
                    'No job found with ID or name "%s". No jobs are completely failed.',
                    options.job,
                )
            else:
                logger.critical(
                    'No job found with ID or name "%s". How about the failed job %s instead?',
                    options.job,
                    suggestion,
                )
            sys.exit(1)
        elif len(hits) > 1:
            # Several hits, maybe only one has failed
            completely_failed_hits = [job for job in hits if job.remainingTryCount == 0]
            if len(completely_failed_hits) == 0:
                logger.critical(
                    'Multiple jobs match "%s" but none are completely failed: %s',
                    options.job,
                    hits,
                )
                sys.exit(1)
            elif len(completely_failed_hits) > 0:
                logger.critical(
                    'Multiple jobs matching "%s" are completely failed: %s',
                    options.job,
                    completely_failed_hits,
                )
                sys.exit(1)
            else:
                # We found one completely failed job, they probably mean that one.
                logger.info(
                    'There are %s jobs matching "%s"; assuming you mean the failed one: %s',
                    options.job,
                    completely_failed_hits[0],
                )
                job_id = completely_failed_hits[0].jobStoreID
        else:
            # We found one job with this name, so they must mean that one
            logger.info('Looked up job named "%s": %s', options.job, hits[0])
            job_id = hits[0].jobStoreID

    if options.printJobInfo:
        # Report on the job

        if isinstance(jobStore, FileJobStore):
            # List all its files if we can
            printContentsOfJobStore(job_store=jobStore, job_id=job_id)
        # Print the job description itself
        job_desc = jobStore.load_job(job_id)
        print(f"Job: {job_desc}")
        pprint.pprint(job_desc.__dict__)
    else:
        # Run the job

        debug_flags = set()
        local_worker_temp_dir = None
        if options.retrieve_task_directory is not None:
            # Pick a directory in it (which may be removed by the worker) as the worker's temp dir.
            local_worker_temp_dir = os.path.join(
                options.retrieve_task_directory, "worker"
            )
            # Make sure it exists
            os.makedirs(local_worker_temp_dir, exist_ok=True)
            # And tell the job to just download files
            debug_flags.add("download_only")
        # We might need to reconstruct a container environment.
        host_and_job_paths: Optional[list[tuple[str, str]]] = None
        # Track if the run succeeded without error
        run_succeeded = False

        logger.info(f"Running the following job locally: {job_id}")
        try:
            workerScript(
                jobStore,
                config,
                job_id,
                job_id,
                redirect_output_to_log_file=False,
                local_worker_temp_dir=local_worker_temp_dir,
                debug_flags=debug_flags,
            )
        except FilesDownloadedStoppingPointReached as e:
            # We asked for the files to be downloaded and now they are.
            assert options.retrieve_task_directory is not None
            if e.host_and_job_paths is not None:
                # Capture the container mapping so we can reconstruct the container environment after we unwind the worker stack.
                host_and_job_paths = e.host_and_job_paths
        else:
            # No error!
            run_succeeded = True

        # Make sure the deferred function manager cleans up and logs its
        # shutdown before we start writing any reports.
        gc.collect()

        if run_succeeded:
            logger.info(f"Successfully ran: {job_id}")

        if host_and_job_paths is not None:
            # We need to make a place that looks like the job paths half of these.

            # Sort by job-side path so we do children before parents, to
            # stop us from accidentally making children inside moutned
            # parents.
            sorted_mounts = sorted(host_and_job_paths, key=lambda t: t[1], reverse=True)

            fake_job_root = os.path.join(options.retrieve_task_directory, "inside")
            os.makedirs(fake_job_root, exist_ok=True)

            for host_path, job_path in sorted_mounts:
                if not os.path.exists(host_path):
                    logger.error(
                        "Job intended to mount %s as %s but it does not exist!",
                        host_path,
                        job_path,
                    )
                    continue
                if not job_path.startswith("/"):
                    logger.error(
                        "Job intended to mount %s as %s but destination is a relative path!",
                        host_path,
                        job_path,
                    )
                    continue
                # Drop the slash because we are building a chroot-ish mini filesystem.
                job_relative_path = job_path[1:]
                if job_relative_path.startswith("/"):
                    # We are having trouble understanding what the job
                    # intended to do. Stop working on this mount.
                    logger.error(
                        "Job intended to mount %s as %s but destination starts with multiple slashes for some reason!",
                        host_path,
                        job_path,
                    )
                    continue
                fake_job_path = os.path.join(fake_job_root, job_relative_path)
                if os.path.exists(fake_job_path):
                    logger.error(
                        "Job intended to mount %s as %s but that location is already mounted!",
                        host_path,
                        job_path,
                    )
                    continue

                logger.info("Job mounted %s as %s", host_path, job_path)

                # Make sure the directory to contain the mount exists.
                fake_job_containing_path = os.path.dirname(fake_job_path)
                os.makedirs(fake_job_containing_path, exist_ok=True)

                top_pathobj = Path(os.path.abspath(options.retrieve_task_directory))
                source_pathobj = Path(host_path)
                if top_pathobj in source_pathobj.parents:
                    # We're linking to a file we already downloaded (probably).
                    # Make a relative symlink so the whole assemblage can move.
                    host_path = os.path.relpath(host_path, fake_job_containing_path)

                # Make a symlink to simulate the mount
                os.symlink(host_path, fake_job_path)

            logger.info("Reconstructed job container filesystem at %s", fake_job_root)

    # TODO: Option to print list of successor jobs
    # TODO: Option to run job within python debugger, allowing step through of arguments
    # idea would be to have option to import pdb and set breakpoint at the start of the user's code



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Generator
import argparse
import logging
import os
from pathlib import Path
import time
from typing import Any

from toil.test.batchSystems import batchSystemTest
from toil.batchSystems.mesos.test import MesosTestSupport
from toil.fileStores import FileID
from toil.job import Job, PromisedRequirement, JobFunctionWrappingJob, Promise
from toil.lib.retry import retry_flaky_test
from toil.test import pneeds_mesos as needs_mesos, pslow as slow

import pytest

log = logging.getLogger(__name__)


class AbstractPromisedRequirementsTest(batchSystemTest.AbstractBatchSystemJobTest):
    """An abstract base class for testing Toil workflows with promised requirements."""

    @slow
    @pytest.mark.slow
    def testConcurrencyDynamic(self, tmp_path: Path) -> None:
        """
        Asserts that promised core resources are allocated properly using a dynamic Toil workflow
        """
        for coresPerJob in self.allocatedCores:
            log.debug(
                "Testing %d cores per job with CPU count %d",
                coresPerJob,
                self.cpuCount,
            )
            tempDir = tmp_path / f"testFiles_{coresPerJob}"
            tempDir.mkdir()
            counterPath = self.getCounterPath(tempDir)

            root = Job.wrapJobFn(
                maxConcurrency,
                self.cpuCount,
                counterPath,
                coresPerJob,
                cores=1,
                memory="1M",
                disk="1M",
            )
            values = Job.Runner.startToil(root, self.getOptions(tempDir))
            maxValue = max(values)
            assert maxValue <= (self.cpuCount // coresPerJob)

    @slow
    @pytest.mark.slow
    @retry_flaky_test(prepare=[])
    def testConcurrencyStatic(self, tmp_path: Path) -> None:
        """
        Asserts that promised core resources are allocated properly using a static DAG
        """
        for coresPerJob in self.allocatedCores:
            log.debug(
                "Testing %d cores per job with CPU count %d",
                coresPerJob,
                self.cpuCount,
            )
            tempDir = tmp_path / f"testFiles_{coresPerJob}"
            tempDir.mkdir()
            counterPath = self.getCounterPath(tempDir)

            root = Job()
            one = Job.wrapFn(getOne, cores=0.1, memory="32M", disk="1M")
            thirtyTwoMb = Job.wrapFn(getThirtyTwoMb, cores=0.1, memory="32M", disk="1M")
            root.addChild(one)
            root.addChild(thirtyTwoMb)
            for _ in range(self.cpuCount):
                root.addFollowOn(
                    Job.wrapFn(
                        batchSystemTest.measureConcurrency,
                        counterPath,
                        cores=PromisedRequirement(lambda x: x * coresPerJob, one.rv()),
                        memory=PromisedRequirement(thirtyTwoMb.rv()),
                        disk="1M",
                    )
                )
            Job.Runner.startToil(root, self.getOptions(tempDir))
            _, maxValue = batchSystemTest.getCounters(counterPath)
            assert maxValue <= (self.cpuCount // coresPerJob)

    def getOptions(self, tempDir: Path, caching: bool = True) -> argparse.Namespace:
        options = super().getOptions(tempDir)
        # defaultCores defaults to 1 - this is coincidentally the core requirement relied upon by this
        # test, so we change defaultCores to 2 to make the test more strict
        options.defaultCores = 2
        options.caching = caching
        return options

    def getCounterPath(self, tempDir: Path) -> Path:
        """
        Returns path to a counter file
        :param tempDir: path to test directory
        :return: path to counter file
        """
        counterPath = tempDir / "counter"
        batchSystemTest.resetCounters(counterPath)
        minValue, maxValue = batchSystemTest.getCounters(counterPath)
        assert (minValue, maxValue) == (0, 0)
        return counterPath

    def testPromisesWithJobStoreFileObjects(
        self, tmp_path: Path, caching: bool = True
    ) -> None:
        """
        Check whether FileID objects are being pickled properly when used as return
        values of functions.  Then ensure that lambdas of promised FileID objects can be
        used to describe the requirements of a subsequent job.  This type of operation will be
        used commonly in Toil scripts.
        :return: None
        """
        file1 = 1024
        file2 = 512
        F1 = Job.wrapJobFn(_writer, file1)
        F2 = Job.wrapJobFn(_writer, file2)
        G = Job.wrapJobFn(
            _follower,
            file1 + file2,
            disk=PromisedRequirement(lambda x, y: x.size + y.size, F1.rv(), F2.rv()),
        )
        F1.addChild(F2)
        F2.addChild(G)

        Job.Runner.startToil(F1, self.getOptions(tmp_path, caching=caching))

    def testPromisesWithNonCachingFileStore(self, tmp_path: Path) -> None:
        self.testPromisesWithJobStoreFileObjects(tmp_path, caching=False)

    @slow
    @pytest.mark.slow
    def testPromiseRequirementRaceStatic(self, tmp_path: Path) -> None:
        """
        Checks for a race condition when using promised requirements and child job functions.
        """
        A = Job.wrapJobFn(logDiskUsage, "A", sleep=5, disk=PromisedRequirement(1024))
        B = Job.wrapJobFn(
            logDiskUsage, "B", disk=PromisedRequirement(lambda x: x + 1024, A.rv())
        )
        A.addChild(B)
        Job.Runner.startToil(A, self.getOptions(tmp_path))


def _writer(job: JobFunctionWrappingJob, fileSize: int) -> FileID:
    """
    Write a local file and return the FileID obtained from running writeGlobalFile on
    it.

    :param job: job
    :param fileSize: Size of the file in bytes
    :returns: the result of writeGlobalFile on a locally created file
    :rtype: job.FileID
    """
    with open(job.fileStore.getLocalTempFileName(), "wb") as fH:
        fH.write(os.urandom(fileSize))
    return job.fileStore.writeGlobalFile(fH.name)


def _follower(job: Job, expectedDisk: Any) -> None:
    """
    This job follows the _writer jobs and ensures the expected disk is used.

    :param job: job
    :param expectedDisk: Expect disk to be used by this job
    :return: None
    """
    assert job.disk == expectedDisk


def maxConcurrency(
    job: Job, cpuCount: int, filename: Path, coresPerJob: int
) -> list[Promise]:
    """
    Returns the max number of concurrent tasks when using a PromisedRequirement instance
    to allocate the number of cores per job.

    :param cpuCount: number of available cpus
    :param filename: path to counter file
    :param coresPerJob: number of cores assigned to each job
    :return: max concurrency value per CPU
    """
    one = job.addChildFn(getOne, cores=0.1, memory="32M", disk="1M")
    thirtyTwoMb = job.addChildFn(getThirtyTwoMb, cores=0.1, memory="32M", disk="1M")

    values = []
    for _ in range(cpuCount):
        value = job.addFollowOnFn(
            batchSystemTest.measureConcurrency,
            filename,
            cores=PromisedRequirement(lambda x: x * coresPerJob, one.rv()),
            memory=PromisedRequirement(thirtyTwoMb.rv()),
            disk="1M",
        ).rv()
        values.append(value)
    return values


def getOne() -> int:
    return 1


def getThirtyTwoMb() -> str:
    return "32M"


def logDiskUsage(job: JobFunctionWrappingJob, funcName: str, sleep: int = 0) -> int:
    """
    Logs the job's disk usage to master and sleeps for specified amount of time.

    :return: job function's disk usage
    """
    diskUsage = job.disk
    job.fileStore.log_to_leader(f"{funcName}: {diskUsage}")
    time.sleep(sleep)
    return diskUsage


class TestSingleMachinePromisedRequirements(AbstractPromisedRequirementsTest):
    """
    Tests against the SingleMachine batch system
    """

    def getBatchSystemName(self) -> str:
        return "single_machine"

    def tearDown(self) -> None:
        pass


@needs_mesos
class TestMesosPromisedRequirements(AbstractPromisedRequirementsTest, MesosTestSupport):
    """
    Tests against the Mesos batch system
    """

    @pytest.fixture(autouse=True)
    def mesos_support(self) -> Generator[None]:
        try:
            self._startMesos(self.cpuCount)
            yield
        finally:
            self._stopMesos()

    def getOptions(self, tempDir: Path, caching: bool = True) -> argparse.Namespace:
        options = super().getOptions(tempDir, caching=caching)
        options.mesos_endpoint = "localhost:5050"
        return options

    def getBatchSystemName(self) -> str:
        return "mesos"



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
import types
from importlib.machinery import SourceFileLoader
from tempfile import NamedTemporaryFile

from setuptools import find_packages, setup

SETUP_DIR = os.path.dirname(__file__)
README = os.path.join(SETUP_DIR, "README.rst")


def get_requirements(extra=None):
    """
    Load the requirements for the given extra.

    Uses the appropriate requirements-extra.txt, or the main requirements.txt
    if no extra is specified.
    """
    filename = f"requirements-{extra}.txt" if extra else "requirements.txt"

    with open(filename) as fp:
        # Parse out as one per line, dropping comments
        return [
            l.split("#")[0].strip() for l in fp.readlines() if l.split("#")[0].strip()
        ]


def run_setup():
    """
    Call setup().

    This function exists so the setup() invocation preceded more internal functionality.
    The `version` module is imported dynamically by import_version() below.
    """
    install_requires = get_requirements()

    extras_require = {}
    # htcondor is not supported by apple
    # this is tricky to conditionally support in 'all' due
    # to how wheels work, so it is not included in all and
    # must be explicitly installed as an extra
    all_reqs = []

    non_htcondor_extras = [
        "aws",
        "cwl",
        "encryption",
        "google",
        "kubernetes",
        "wdl",
        "server",
    ]
    for extra in non_htcondor_extras:
        extras_require[extra] = get_requirements(extra)
        all_reqs += extras_require[extra]
    # We exclude htcondor from "all" because it can't be on Mac
    extras_require['htcondor:sys_platform!="darwin"'] = get_requirements("htcondor")
    extras_require["mesos"] = get_requirements("mesos")
    all_reqs += get_requirements("mesos")
    extras_require["all"] = all_reqs
    setup(
        name="toil",
        version=version.distVersion,
        long_description=open(README).read(),
        long_description_content_type="text/x-rst",
        description="Pipeline management software for clusters.",
        author="Benedict Paten and the Toil community",
        author_email="toil-community@googlegroups.com",
        url="https://github.com/DataBiosphere/toil",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "Intended Audience :: Healthcare Industry",
            "License :: OSI Approved :: Apache Software License",
            "Natural Language :: English",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: POSIX",
            "Operating System :: POSIX :: Linux",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Programming Language :: Python :: 3.13",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
            "Topic :: Scientific/Engineering :: Astronomy",
            "Topic :: Scientific/Engineering :: Atmospheric Science",
            "Topic :: Scientific/Engineering :: Information Analysis",
            "Topic :: Scientific/Engineering :: Medical Science Apps.",
            "Topic :: System :: Distributed Computing",
            "Topic :: Utilities",
        ],
        license="Apache License v2.0",
        python_requires=">=3.9",
        install_requires=install_requires,
        extras_require=extras_require,
        package_dir={"": "src"},
        packages=find_packages(where="src"),
        include_package_data=True,
        # Unfortunately, the names of the entry points are hard-coded elsewhere in the code base so
        # you can't just change them here. Luckily, most of them are pretty unique strings, and thus
        # easy to search for.
        entry_points={
            "console_scripts": [
                "toil = toil.utils.toilMain:main",
                "_toil_worker = toil.worker:main",
                "cwltoil = toil.cwl.cwltoil:cwltoil_was_removed [cwl]",
                "toil-cwl-runner = toil.cwl.cwltoil:main [cwl]",
                "toil-wdl-runner = toil.wdl.wdltoil:main [wdl]",
                "toil-wes-cwl-runner = toil.server.cli.wes_cwl_runner:main [server]",
                "_toil_mesos_executor = toil.batchSystems.mesos.executor:main [mesos]",
                "_toil_contained_executor = toil.batchSystems.contained_executor:executor",
            ]
        },
    )


def import_version():
    """Return the module object for src/toil/version.py, generate from the template if required."""
    if not os.path.exists("src/toil/version.py"):
        for req in get_requirements("cwl"):
            # Determine cwltool version from requirements file
            if req.startswith("cwltool=="):
                cwltool_version = req[len("cwltool==") :]
                break
        # Use the template to generate src/toil/version.py

        # First make sure we can find it (in case we're happening during an
        # install where the source tree isn't on the path already).
        # See <https://github.com/pypa/setuptools/discussions/3909#discussioncomment-5718455>
        sys.path.append(os.path.dirname(__file__))

        import version_template

        with NamedTemporaryFile(
            mode="w", dir="src/toil", prefix="version.py.", delete=False
        ) as f:
            f.write(
                version_template.expand_(
                    others={
                        # expose the dependency versions that we may need to access in Toil
                        "cwltool_version": cwltool_version,
                    }
                )
            )
        os.rename(f.name, "src/toil/version.py")

    loader = SourceFileLoader("toil.version", "src/toil/version.py")
    mod = types.ModuleType(loader.name)
    loader.exec_module(mod)
    return mod


version = import_version()
run_setup()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from toil.common import Toil, ToilContextManagerException
from toil.job import Job
from toil.test import ToilTest, get_temp_file, slow


@slow
class ToilContextManagerTest(ToilTest):
    def setUp(self):
        self.exportPath = get_temp_file()

    def tearDown(self):
        os.remove(self.exportPath)

    def testContextManger(self):
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        options.logLevel = "INFO"
        with Toil(options) as toil:
            toil.start(HelloWorld())

    def testNoContextManger(self):
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        options.logLevel = "INFO"
        toil = Toil(options)
        self.assertRaises(ToilContextManagerException, toil.start, HelloWorld())

    def testExportAfterFailedExport(self):
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        try:
            with Toil(options) as toil:
                _ = toil.start(HelloWorld())
                # oh no, an error! :(
                raise RuntimeError(
                    "we died after workflow completion but before our export finished"
                )
        except RuntimeError:
            pass

        options.restart = True
        with Toil(options) as toil:
            fileID = toil.restart()
            print(fileID)
            # Hopefully the error didn't cause us to lose all our work!
            toil.exportFile(fileID, "file://" + self.exportPath)
        with open(self.exportPath) as f:
            # The file should have all our content
            self.assertEqual(f.read(), "Hello, World!")


class HelloWorld(Job):
    def __init__(self):
        Job.__init__(self, memory=100000, disk="1M")

    def run(self, fileStore):
        fileID = self.addChildJobFn(childFn, memory="1M", disk="1M").rv()
        return self.addFollowOn(FollowOn(fileID)).rv()


def childFn(job):
    with job.fileStore.writeGlobalFileStream() as (fH, fileID):
        fH.write(b"Hello, World!")
        return fileID


class FollowOn(Job):
    def __init__(self, fileId):
        Job.__init__(self)
        self.fileId = fileId

    def run(self, fileStore):
        tempDir = fileStore.getLocalTempDir()
        tempFilePath = "/".join([tempDir, "LocalCopy"])
        with fileStore.readGlobalFileStream(self.fileId) as globalFile:
            with open(tempFilePath, "wb") as localFile:
                localFile.write(globalFile.read())
        return self.fileId



from toil.common import Toil
from toil.job import Job


def helloWorld(job, message):
    job.log(f"Hello world, I have a message: {message}")


if __name__ == "__main__":
    parser = Job.Runner.getDefaultArgumentParser()
    options = parser.parse_args()
    options.logLevel = "INFO"
    options.clean = "always"

    j1 = Job.wrapJobFn(helloWorld, "first")
    j2 = Job.wrapJobFn(helloWorld, "second or third")
    j3 = Job.wrapJobFn(helloWorld, "second or third")
    j4 = Job.wrapJobFn(helloWorld, "last")

    j1.addChild(j2)
    j1.addChild(j3)
    j1.addFollowOn(j4)

    with Toil(options) as toil:
        toil.start(j1)
